"""Task of OTX Detection using mmdetection training backend."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import glob
import io
import os
import time
from contextlib import nullcontext
from copy import deepcopy
from functools import partial
from typing import Any, Dict, Optional, Union

import torch
from mmcv.runner import wrap_fp16_model
from mmcv.utils import Config, ConfigDict, get_git_hash
from mmdet import __version__
from mmdet.apis import single_gpu_test, train_detector
from mmdet.datasets import build_dataloader, build_dataset, replace_ImageToTensor
from mmdet.models.detectors import TwoStageDetector
from mmdet.utils import collect_env

from otx.algorithms.common.adapters.mmcv.hooks.recording_forward_hook import (
    ActivationMapHook,
    BaseRecordingForwardHook,
    EigenCamHook,
    FeatureVectorHook,
)
from otx.algorithms.common.adapters.mmcv.utils import (
    adapt_batch_size,
    build_data_parallel,
    patch_data_pipeline,
    patch_from_hyperparams,
)
from otx.algorithms.common.adapters.mmcv.utils.config_utils import (
    MPAConfig,
    update_or_add_custom_hook,
)
from otx.algorithms.common.configs.training_base import TrainType
from otx.algorithms.common.tasks.nncf_task import NNCFBaseTask
from otx.algorithms.common.utils import set_random_seed
from otx.algorithms.common.utils.data import get_dataset
from otx.algorithms.common.utils.logger import get_logger
from otx.algorithms.detection.adapters.mmdet.configurer import (
    DetectionConfigurer,
    IncrDetectionConfigurer,
    SemiSLDetectionConfigurer,
)
from otx.algorithms.detection.adapters.mmdet.datasets import ImageTilingDataset
from otx.algorithms.detection.adapters.mmdet.hooks.det_class_probability_map_hook import (
    DetClassProbabilityMapHook,
)
from otx.algorithms.detection.adapters.mmdet.utils import (
    patch_input_preprocessing,
    patch_input_shape,
    patch_ir_scale_factor,
    patch_tiling,
)
from otx.algorithms.detection.adapters.mmdet.utils.builder import build_detector
from otx.algorithms.detection.adapters.mmdet.utils.config_utils import (
    should_cluster_anchors,
)
from otx.algorithms.detection.adapters.mmdet.utils.exporter import DetectionExporter
from otx.algorithms.detection.task import OTXDetectionTask
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import ids_to_strings
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.explain_parameters import ExplainParameters
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import (
    ModelEntity,
    ModelPrecision,
)
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.core.data import caching

logger = get_logger()

# TODO Remove unnecessary pylint disable
# pylint: disable=too-many-lines


class MMDetectionTask(OTXDetectionTask):
    """Task class for OTX detection using mmdetection training backend."""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, task_environment: TaskEnvironment, output_path: Optional[str] = None):
        super().__init__(task_environment, output_path)
        self._data_cfg: Optional[Config] = None
        self._recipe_cfg: Optional[Config] = None

    # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    def _init_task(self, dataset: Optional[DatasetEntity] = None, export: bool = False):  # noqa
        """Initialize task."""
        self._recipe_cfg = MPAConfig.fromfile(os.path.join(self._model_dir, "model.py"))
        self._recipe_cfg.domain = self._task_type.domain
        self._config = self._recipe_cfg

        set_random_seed(self._recipe_cfg.get("seed", 5), logger, self._recipe_cfg.get("deterministic", False))

        # Belows may go to the configure function
        patch_data_pipeline(self._recipe_cfg, self.data_pipeline_path)

        # Patch tiling parameters
        patch_tiling(self._recipe_cfg, self._hyperparams, dataset)

        if not export:
            patch_from_hyperparams(self._recipe_cfg, self._hyperparams)

        if "custom_hooks" in self.override_configs:
            override_custom_hooks = self.override_configs.pop("custom_hooks")
            for override_custom_hook in override_custom_hooks:
                update_or_add_custom_hook(self._recipe_cfg, ConfigDict(override_custom_hook))
        if len(self.override_configs) > 0:
            logger.info(f"before override configs merging = {self._recipe_cfg}")
            self._recipe_cfg.merge_from_dict(self.override_configs)
            logger.info(f"after override configs merging = {self._recipe_cfg}")

        # add Cancel training hook
        update_or_add_custom_hook(
            self._recipe_cfg,
            ConfigDict(type="CancelInterfaceHook", init_callback=self.on_hook_initialized),
        )
        if self._time_monitor is not None:
            update_or_add_custom_hook(
                self._recipe_cfg,
                ConfigDict(
                    type="OTXProgressHook",
                    time_monitor=self._time_monitor,
                    verbose=True,
                    priority=71,
                ),
            )
        self._recipe_cfg.log_config.hooks.append({"type": "OTXLoggerHook", "curves": self._learning_curves})

        # Update recipe with caching modules
        self._update_caching_modules(self._recipe_cfg.data)

        logger.info("initialized.")

    def build_model(
        self,
        cfg: Config,
        fp16: bool = False,
        **kwargs,
    ) -> torch.nn.Module:
        """Build model from model_builder."""
        model_builder = getattr(self, "model_builder", build_detector)
        model = model_builder(cfg, **kwargs)
        if bool(fp16):
            wrap_fp16_model(model)
        return model

    # pylint: disable=too-many-arguments
    def configure(
        self,
        training=True,
        subset="train",
        ir_options=None,
    ):
        """Patch mmcv configs for OTX detection settings."""

        # deepcopy all configs to make sure
        # changes under MPA and below does not take an effect to OTX for clear distinction
        recipe_cfg = deepcopy(self._recipe_cfg)
        data_cfg = deepcopy(self._data_cfg)
        assert recipe_cfg is not None, "'recipe_cfg' is not initialized."

        if self._data_cfg is not None:
            data_classes = [label.name for label in self._labels]
        else:
            data_classes = None
        model_classes = [label.name for label in self._model_label_schema]

        recipe_cfg.work_dir = self._output_path
        recipe_cfg.resume = self._resume

        if self._train_type == TrainType.Incremental:
            configurer = IncrDetectionConfigurer()
        elif self._train_type == TrainType.Semisupervised:
            configurer = SemiSLDetectionConfigurer()
        else:
            configurer = DetectionConfigurer()
        cfg = configurer.configure(
            recipe_cfg, self._model_ckpt, data_cfg, training, subset, ir_options, data_classes, model_classes
        )
        self._config = cfg
        return cfg

    # pylint: disable=too-many-branches, too-many-statements
    def _train_model(
        self,
        dataset: DatasetEntity,
    ):
        """Train function in MMDetectionTask."""
        logger.info("init data cfg.")
        self._data_cfg = ConfigDict(data=ConfigDict())

        for cfg_key, subset in zip(
            ["train", "val", "unlabeled"],
            [Subset.TRAINING, Subset.VALIDATION, Subset.UNLABELED],
        ):
            subset = get_dataset(dataset, subset)
            if subset and self._data_cfg is not None:
                self._data_cfg.data[cfg_key] = ConfigDict(
                    otx_dataset=subset,
                    labels=self._labels,
                )

        self._is_training = True

        self._init_task(dataset)

        cfg = self.configure(True, "train", None)
        logger.info("train!")

        timestamp = time.strftime("%Y%m%d_%H%M%S", time.localtime())

        # Environment
        logger.info(f"cfg.gpu_ids = {cfg.gpu_ids}, distributed = {cfg.distributed}")
        env_info_dict = collect_env()
        env_info = "\n".join([(f"{k}: {v}") for k, v in env_info_dict.items()])
        dash_line = "-" * 60 + "\n"
        logger.info(f"Environment info:\n{dash_line}{env_info}\n{dash_line}")

        # Data
        datasets = [build_dataset(cfg.data.train)]

        # FIXME: Currently detection do not support multi batch evaluation. This will be fixed
        if "val" in cfg.data:
            cfg.data.val_dataloader["samples_per_gpu"] = 1

        # TODO. This should be moved to configurer
        # TODO. Anchor clustering should be checked
        # if hasattr(cfg, "hparams"):
        #     if cfg.hparams.get("adaptive_anchor", False):
        #         num_ratios = cfg.hparams.get("num_anchor_ratios", 5)
        #         proposal_ratio = extract_anchor_ratio(datasets[0], num_ratios)
        #         self.configure_anchor(cfg, proposal_ratio)

        # Target classes
        if "task_adapt" in cfg:
            target_classes = cfg.task_adapt.get("final", [])
        else:
            target_classes = datasets[0].CLASSES

        # Metadata
        meta = dict()
        meta["env_info"] = env_info
        # meta['config'] = cfg.pretty_text
        meta["seed"] = cfg.seed
        meta["exp_name"] = cfg.work_dir
        if cfg.checkpoint_config is not None:
            cfg.checkpoint_config.meta = dict(
                mmdet_version=__version__ + get_git_hash()[:7],
                CLASSES=target_classes,
            )
            # if "proposal_ratio" in locals():
            #     cfg.checkpoint_config.meta.update({"anchor_ratio": proposal_ratio})

        # Model
        model = self.build_model(cfg, fp16=cfg.get("fp16", False))
        model.train()
        model.CLASSES = target_classes

        if cfg.distributed:
            torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)

        validate = bool(cfg.data.get("val", None))

        if self._hyperparams.learning_parameters.auto_decrease_batch_size:
            train_func = partial(train_detector, meta=deepcopy(meta), model=deepcopy(model), distributed=False)
            adapt_batch_size(train_func, cfg, datasets, isinstance(self, NNCFBaseTask))  # nncf needs eval hooks

        train_detector(
            model,
            datasets,
            cfg,
            distributed=cfg.distributed,
            validate=validate,
            timestamp=timestamp,
            meta=meta,
        )

        # Save outputs
        output_ckpt_path = os.path.join(cfg.work_dir, "latest.pth")
        best_ckpt_path = glob.glob(os.path.join(cfg.work_dir, "best_*.pth"))
        if len(best_ckpt_path) > 0:
            output_ckpt_path = best_ckpt_path[0]
        return dict(
            final_ckpt=output_ckpt_path,
        )

    def _infer_model(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ):
        """Main infer function."""
        self._data_cfg = ConfigDict(
            data=ConfigDict(
                train=ConfigDict(
                    otx_dataset=None,
                    labels=self._labels,
                ),
                test=ConfigDict(
                    otx_dataset=dataset,
                    labels=self._labels,
                ),
            )
        )

        dump_features = True
        dump_saliency_map = not inference_parameters.is_evaluation if inference_parameters else True

        self._init_task(dataset)

        cfg = self.configure(False, "test", None)
        logger.info("infer!")

        samples_per_gpu = 1

        # Data loader
        mm_dataset = build_dataset(cfg.data.test)
        dataloader = build_dataloader(
            mm_dataset,
            samples_per_gpu=samples_per_gpu,
            workers_per_gpu=cfg.data.test_dataloader.get("workers_per_gpu", 0),
            num_gpus=len(cfg.gpu_ids),
            dist=cfg.distributed,
            seed=cfg.get("seed", None),
            shuffle=False,
        )

        # Target classes
        if "task_adapt" in cfg:
            target_classes = cfg.task_adapt.final
            if len(target_classes) < 1:
                raise KeyError(
                    f"target_classes={target_classes} is empty check the metadata from model ckpt or recipe "
                    "configuration"
                )
        else:
            target_classes = mm_dataset.CLASSES

        # Model
        model = self.build_model(cfg, fp16=cfg.get("fp16", False))
        model.CLASSES = target_classes
        model.eval()
        feature_model = model.model_t if self._train_type == TrainType.Semisupervised else model
        model = build_data_parallel(model, cfg, distributed=False)

        # InferenceProgressCallback (Time Monitor enable into Infer task)
        time_monitor = None
        if cfg.get("custom_hooks", None):
            time_monitor = [hook.time_monitor for hook in cfg.custom_hooks if hook.type == "OTXProgressHook"]
            time_monitor = time_monitor[0] if time_monitor else None
        if time_monitor is not None:

            # pylint: disable=unused-argument
            def pre_hook(module, inp):
                time_monitor.on_test_batch_begin(None, None)

            def hook(module, inp, outp):
                time_monitor.on_test_batch_end(None, None)

            model.register_forward_pre_hook(pre_hook)
            model.register_forward_hook(hook)

        # Class-wise Saliency map for Single-Stage Detector, otherwise use class-ignore saliency map.
        if not dump_saliency_map:
            saliency_hook: Union[nullcontext, BaseRecordingForwardHook] = nullcontext()
        else:
            raw_model = feature_model
            if raw_model.__class__.__name__ == "NNCFNetwork":
                raw_model = raw_model.get_nncf_wrapped_model()
            if isinstance(raw_model, TwoStageDetector):
                saliency_hook = ActivationMapHook(feature_model)
            else:
                saliency_hook = DetClassProbabilityMapHook(feature_model)

        if not dump_features:
            feature_vector_hook: Union[nullcontext, BaseRecordingForwardHook] = nullcontext()
        else:
            feature_vector_hook = FeatureVectorHook(feature_model)

        eval_predictions = []
        # pylint: disable=no-member
        with feature_vector_hook:
            with saliency_hook:
                eval_predictions = single_gpu_test(model, dataloader)
                if isinstance(feature_vector_hook, nullcontext):
                    feature_vectors = [None] * len(mm_dataset)
                else:
                    feature_vectors = feature_vector_hook.records
                if isinstance(saliency_hook, nullcontext):
                    saliency_maps = [None] * len(mm_dataset)
                else:
                    saliency_maps = saliency_hook.records

        for key in ["interval", "tmpdir", "start", "gpu_collect", "save_best", "rule", "dynamic_intervals"]:
            cfg.evaluation.pop(key, None)

        metric = None
        if inference_parameters and inference_parameters.is_evaluation:
            metric = mm_dataset.evaluate(eval_predictions, **cfg.evaluation)
            metric = metric["mAP"] if isinstance(cfg.evaluation.metric, list) else metric[cfg.evaluation.metric]

        # Check and unwrap ImageTilingDataset object from TaskAdaptEvalDataset
        while hasattr(mm_dataset, "dataset") and not isinstance(mm_dataset, ImageTilingDataset):
            mm_dataset = mm_dataset.dataset

        if isinstance(mm_dataset, ImageTilingDataset):
            feature_vectors = [feature_vectors[i] for i in range(mm_dataset.num_samples)]
            saliency_maps = [saliency_maps[i] for i in range(mm_dataset.num_samples)]
            if not mm_dataset.merged_results:
                eval_predictions = mm_dataset.merge(eval_predictions)
            else:
                eval_predictions = mm_dataset.merged_results

        assert len(eval_predictions) == len(feature_vectors) == len(saliency_maps), (
            "Number of elements should be the same, however, number of outputs are "
            f"{len(eval_predictions)}, {len(feature_vectors)}, and {len(saliency_maps)}"
        )

        results = dict(
            outputs=dict(
                classes=target_classes,
                detections=eval_predictions,
                metric=metric,
                feature_vectors=feature_vectors,
                saliency_maps=saliency_maps,
            )
        )

        # TODO: InferenceProgressCallback register
        output = results["outputs"]
        metric = output["metric"]
        predictions = output["detections"]
        assert len(output["detections"]) == len(output["feature_vectors"]) == len(output["saliency_maps"]), (
            "Number of elements should be the same, however, number of outputs are "
            f"{len(output['detections'])}, {len(output['feature_vectors'])}, and {len(output['saliency_maps'])}"
        )
        prediction_results = zip(predictions, output["feature_vectors"], output["saliency_maps"])
        return prediction_results, metric

    # pylint: disable=too-many-statements
    def _export_model(
        self,
        precision: ModelPrecision,
        dump_features: bool,
    ):
        """Main export function of OTX MMDetection Task."""
        self._init_task(export=True)

        cfg = self.configure(False, "test", None)

        self._precision[0] = precision
        export_options: Dict[str, Any] = {}
        export_options["deploy_cfg"] = self._init_deploy_cfg(cfg)
        if export_options.get("precision", None) is None:
            assert len(self._precision) == 1
            export_options["precision"] = str(self._precision[0])

        export_options["deploy_cfg"]["dump_features"] = dump_features
        if dump_features:
            output_names = export_options["deploy_cfg"]["ir_config"]["output_names"]
            if "feature_vector" not in output_names:
                output_names.append("feature_vector")
            if export_options["deploy_cfg"]["codebase_config"]["task"] != "Segmentation":
                if "saliency_map" not in output_names:
                    output_names.append("saliency_map")
        export_options["model_builder"] = getattr(self, "model_builder", build_detector)

        if self._precision[0] == ModelPrecision.FP16:
            export_options["deploy_cfg"]["backend_config"]["mo_options"]["flags"].append("--compress_to_fp16")

        exporter = DetectionExporter()
        results = exporter.run(
            cfg,
            **export_options,
        )

        return results

    def _explain_model(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[ExplainParameters] = None,
    ) -> Dict[str, Any]:
        """Main explain function of MMDetectionTask."""

        explainer_hook_selector = {
            "classwisesaliencymap": DetClassProbabilityMapHook,
            "eigencam": EigenCamHook,
            "activationmap": ActivationMapHook,
        }
        self._data_cfg = ConfigDict(
            data=ConfigDict(
                train=ConfigDict(
                    otx_dataset=None,
                    labels=self._labels,
                ),
                test=ConfigDict(
                    otx_dataset=dataset,
                    labels=self._labels,
                ),
            )
        )

        self._init_task()

        cfg = self.configure(False, "test", None)

        samples_per_gpu = cfg.data.test_dataloader.get("samples_per_gpu", 1)
        if samples_per_gpu > 1:
            # Replace 'ImageToTensor' to 'DefaultFormatBundle'
            cfg.data.test.pipeline = replace_ImageToTensor(cfg.data.test.pipeline)

        # Data loader
        mm_dataset = build_dataset(cfg.data.test)
        dataloader = build_dataloader(
            mm_dataset,
            samples_per_gpu=cfg.data.get("samples_per_gpu", 1),
            workers_per_gpu=cfg.data.get("workers_per_gpu", 0),
            num_gpus=len(cfg.gpu_ids),
            dist=cfg.distributed,
            seed=cfg.get("seed", None),
            shuffle=False,
        )

        # Target classes
        if "task_adapt" in cfg:
            target_classes = cfg.task_adapt.final
            if len(target_classes) < 1:
                raise KeyError(
                    f"target_classes={target_classes} is empty check the metadata from model ckpt or recipe "
                    "configuration"
                )
        else:
            target_classes = mm_dataset.CLASSES

        # TODO: Check Inference FP16 Support
        model = self.build_model(cfg, fp16=cfg.get("fp16", False))
        model.CLASSES = target_classes
        model.eval()
        feature_model = model.model_t if self._train_type == TrainType.Semisupervised else model
        model = build_data_parallel(model, cfg, distributed=False)

        # InferenceProgressCallback (Time Monitor enable into Infer task)
        time_monitor = None
        if cfg.get("custom_hooks", None):
            time_monitor = [hook.time_monitor for hook in cfg.custom_hooks if hook.type == "OTXProgressHook"]
            time_monitor = time_monitor[0] if time_monitor else None
        if time_monitor is not None:

            # pylint: disable=unused-argument
            def pre_hook(module, inp):
                time_monitor.on_test_batch_begin(None, None)

            def hook(module, inp, outp):
                time_monitor.on_test_batch_end(None, None)

            model.register_forward_pre_hook(pre_hook)
            model.register_forward_hook(hook)

        explainer = explain_parameters.explainer if explain_parameters else None
        if explainer is not None:
            explainer_hook = explainer_hook_selector.get(explainer.lower(), None)
        else:
            explainer_hook = None
        if explainer_hook is None:
            raise NotImplementedError(f"Explainer algorithm {explainer} not supported!")
        logger.info(f"Explainer algorithm: {explainer}")

        # Class-wise Saliency map for Single-Stage Detector, otherwise use class-ignore saliency map.
        eval_predictions = []
        with explainer_hook(feature_model) as saliency_hook:
            for data in dataloader:
                with torch.no_grad():
                    result = model(return_loss=False, rescale=True, **data)
                eval_predictions.extend(result)
            saliency_maps = saliency_hook.records

        # Check and unwrap ImageTilingDataset object from TaskAdaptEvalDataset
        while hasattr(mm_dataset, "dataset") and not isinstance(mm_dataset, ImageTilingDataset):
            mm_dataset = mm_dataset.dataset

        # In the tiling case, select the first images which is map of the entire image
        if isinstance(mm_dataset, ImageTilingDataset):
            saliency_maps = [saliency_maps[i] for i in range(mm_dataset.num_samples)]

        outputs = dict(detections=eval_predictions, saliency_maps=saliency_maps)
        return outputs

    # This should be removed
    def update_override_configurations(self, config):
        """Update override_configs."""
        logger.info(f"update override config with: {config}")
        config = ConfigDict(**config)
        self.override_configs.update(config)

    # This should moved somewhere
    def _init_deploy_cfg(self, cfg) -> Union[Config, None]:
        base_dir = os.path.abspath(os.path.dirname(self._task_environment.model_template.model_template_path))
        if self._hyperparams.tiling_parameters.enable_tile_classifier:
            deploy_cfg_path = os.path.join(base_dir, "deployment_tile_classifier.py")
        else:
            deploy_cfg_path = os.path.join(base_dir, "deployment.py")
        deploy_cfg = None
        if os.path.exists(deploy_cfg_path):
            deploy_cfg = MPAConfig.fromfile(deploy_cfg_path)

            patch_input_preprocessing(cfg, deploy_cfg)
            if not deploy_cfg.backend_config.get("model_inputs", []):
                patch_input_shape(cfg, deploy_cfg)
            patch_ir_scale_factor(deploy_cfg, self._hyperparams)

        return deploy_cfg

    def save_model(self, output_model: ModelEntity):
        """Save best model weights in DetectionTrainTask."""
        logger.info("called save_model")
        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}
        model_ckpt = torch.load(self._model_ckpt)
        modelinfo = {
            "model": model_ckpt,
            "config": hyperparams_str,
            "labels": labels,
            "confidence_threshold": self.confidence_threshold,
            "VERSION": 1,
        }
        if self._recipe_cfg is not None and should_cluster_anchors(self._recipe_cfg):
            modelinfo["anchors"] = {}
            self._update_anchors(modelinfo["anchors"], self._recipe_cfg.model.bbox_head.anchor_generator)

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        output_model.precision = self._precision

    @staticmethod
    def _update_anchors(origin, new):
        logger.info("Updating anchors")
        origin["heights"] = new["heights"]
        origin["widths"] = new["widths"]

    # These need to be moved somewhere
    def _update_caching_modules(self, data_cfg: Config) -> None:
        def _find_max_num_workers(cfg: dict):
            num_workers = [0]
            for key, value in cfg.items():
                if key == "workers_per_gpu" and isinstance(value, int):
                    num_workers += [value]
                elif isinstance(value, dict):
                    num_workers += [_find_max_num_workers(value)]

            return max(num_workers)

        def _get_mem_cache_size():
            if not hasattr(self._hyperparams.algo_backend, "mem_cache_size"):
                return 0

            return self._hyperparams.algo_backend.mem_cache_size

        max_num_workers = _find_max_num_workers(data_cfg)
        mem_cache_size = _get_mem_cache_size()

        mode = "multiprocessing" if max_num_workers > 0 else "singleprocessing"
        caching.MemCacheHandlerSingleton.create(mode, mem_cache_size)

        update_or_add_custom_hook(
            self._recipe_cfg,
            ConfigDict(type="MemCacheHook", priority="VERY_LOW"),
        )



"""Collection of load pipelines for segmentation task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import Any, Dict

from mmseg.datasets.builder import PIPELINES

import otx.core.data.pipelines.load_image_from_otx_dataset as load_image_base
from otx.algorithms.segmentation.adapters.mmseg.datasets.dataset import (
    get_annotation_mmseg_format,
)


# pylint: disable=too-many-instance-attributes, too-many-arguments
@PIPELINES.register_module()
class LoadImageFromOTXDataset(load_image_base.LoadImageFromOTXDataset):
    """Pipeline element that loads an image from a OTX Dataset on the fly."""


@PIPELINES.register_module()
class LoadAnnotationFromOTXDataset:
    """Pipeline element that loads an annotation from a OTX Dataset on the fly.

    Expected entries in the 'results' dict that should be passed to this pipeline element are:
        results['dataset_item']: dataset_item from which to load the annotation
        results['ann_info']['label_list']: list of all labels in the project

    """

    def __init__(self):
        pass

    def __call__(self, results: Dict[str, Any]):
        """Callback function of LoadAnnotationFromOTXDataset."""
        dataset_item = results["dataset_item"]
        labels = results["ann_info"]["labels"]

        ann_info = get_annotation_mmseg_format(dataset_item, labels)

        results["gt_semantic_seg"] = ann_info["gt_semantic_seg"]
        results["seg_fields"].append("gt_semantic_seg")

        return results



"""Algorithm to find a proper batch size which is fit to current GPU device."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from typing import Callable

import torch

from otx.algorithms.common.utils.logger import get_logger

logger = get_logger()


def adapt_batch_size(train_func: Callable[[int], None], current_bs: int, trainset_size: int) -> int:
    """Decrease batch size if default batch size isn't fit to current GPU device.

    Args:
        train_func (Callable[[int], None]): training function with single arugment to set batch size
        current_bs (int): default batch size
        trainset_size (int): training dataset size

    Returns:
        int: proper batch size possibly decreased as default value isn't fit
    """
    if current_bs <= 0:
        raise ValueError("Batch size should be bigger than 0.")
    if trainset_size <= 0:
        raise ValueError("train data set size should be bigger than 0.")

    if trainset_size < current_bs:
        current_bs = trainset_size

    available_bs = 0
    lowest_unavailable_bs = current_bs + 2
    _, total_mem = torch.cuda.mem_get_info()

    def _get_even_center_val(val1, val2):
        ret = (val1 + val2) // 2
        if ret % 2 == 1:
            ret += 1
        return ret

    while True:
        cuda_oom = False
        torch.cuda.reset_max_memory_allocated(device=None)

        try:
            train_func(current_bs)
        except RuntimeError as e:
            if str(e).startswith("CUDA out of memory."):
                cuda_oom = True
            else:
                raise e

        gpu_memory_usage = torch.cuda.max_memory_allocated(device=None) / total_mem
        logger.debug(
            f"Adapting Batch size => bs : {current_bs}, CUDA_OOM : {cuda_oom}, GPU memory usage : {gpu_memory_usage}%"
        )

        # If GPU memory usage is too close to limit, CUDA OOM can be raised during training
        if cuda_oom or torch.cuda.max_memory_allocated(device=None) >= total_mem * 0.85:
            if current_bs < lowest_unavailable_bs:
                lowest_unavailable_bs = current_bs
            current_bs = _get_even_center_val(current_bs, available_bs)
        else:
            available_bs = current_bs
            current_bs = _get_even_center_val(current_bs, lowest_unavailable_bs)

        torch.cuda.empty_cache()
        if lowest_unavailable_bs - available_bs <= 2:
            break

    if available_bs == 0:
        raise RuntimeError("Current device can't train model even with 2!")

    return available_bs



# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import numpy as np
import pytest

from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.usecases.evaluation.basic_operations import (
    divide_arrays_with_possible_zeros,
    get_intersections_and_cardinalities,
    intersection_box,
    intersection_over_union,
    precision_per_class,
    recall_per_class,
)
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestBasicOperationsFunctions:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_get_intersections_and_cardinalities(self):
        """
        <b>Description:</b>
        Check "get_intersections_and_cardinalities" function

        <b>Input data:</b>
        "references" masks array, "predictions" masks array, "labels" list of "LabelEntity" class objects

        <b>Expected results:</b>
        Test passes if tuple returned by "get_intersections_and_cardinalities" function is equal to expected
        """
        equal_array = np.array([(1, 3, 1, 0), (2, 0, 2, 1), (3, 1, 0, 2), (0, 1, 1, 0)])
        other_equal_array = np.array([(2, 1, 1, 0), (1, 0, 2, 0), (2, 1, 0, 1)])
        unequal_reference_array = np.array([(1, 2, 3), (3, 0, 1), (1, 1, 0)])
        unequal_predictions_array = np.array([(0, 2, 3), (3, 0, 2), (1, 0, 1)])
        label_for_first_intersection = LabelEntity(name="label_for_intersection_1", domain=Domain.DETECTION)
        label_for_second_intersection = LabelEntity(name="label_for_intersection_2", domain=Domain.DETECTION)
        label_for_third_intersection = LabelEntity(name="label_for_intersection_3", domain=Domain.DETECTION)
        non_assigned_label = LabelEntity(name="non_assigned", domain=Domain.DETECTION)
        labels = [
            label_for_first_intersection,
            label_for_second_intersection,
            label_for_third_intersection,
            non_assigned_label,
        ]
        intersections_and_cardinalities = get_intersections_and_cardinalities(
            references=[equal_array, unequal_reference_array, other_equal_array],
            predictions=[equal_array, unequal_predictions_array, other_equal_array],
            labels=labels,
        )
        # Checking intersections
        intersections = intersections_and_cardinalities[0]
        assert len(intersections) == 5
        assert intersections.get(label_for_first_intersection) == 12
        assert intersections.get(label_for_second_intersection) == 7
        assert intersections.get(label_for_third_intersection) == 4
        assert intersections.get(non_assigned_label) == 0
        assert intersections.get(None) == 23
        # Checking cardinalities
        cardinalities = intersections_and_cardinalities[1]
        assert len(cardinalities) == 5
        assert cardinalities.get(label_for_first_intersection) == 28
        assert cardinalities.get(label_for_second_intersection) == 15
        assert cardinalities.get(label_for_third_intersection) == 8
        assert cardinalities.get(non_assigned_label) == 0
        assert cardinalities.get(None) == 51

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_intersection_box(self):
        """
        <b>Description:</b>
        Check "intersection_box" function

        <b>Input data:</b>
        "box_1" and "box_2" Rectangle-class objects

        <b>Expected results:</b>
        Test passes if list returned by "intersection_box" function is equal to expected

        <b>Steps</b>
        1. Check list returned by "intersection_box" function for boxes that intersect in two points
        2. Check list returned by "intersection_box" function for boxes that intersect in one point
        3. Check list returned by "intersection_box" function for boxes that intersect by one side
        4. Check list returned by "intersection_box" function when one of boxes completely overlaps other
        5. Check list returned by "intersection_box" function for boxes that not intersect
        """
        box_1 = Rectangle(x1=0.1, y1=0.1, x2=0.3, y2=0.4)
        # Checking list returned by "intersection_box" for boxes that intersect in two points
        box_2 = Rectangle(x1=0.2, y1=0.2, x2=0.5, y2=0.5)
        assert intersection_box(box_1, box_2) == [0.2, 0.2, 0.3, 0.4]
        # Checking list returned by "intersection_box" for boxes that intersect in one point
        box_2 = Rectangle(x1=0.3, y1=0.4, x2=0.5, y2=0.5)
        assert not intersection_box(box_1, box_2)
        # Checking list returned by "intersection_box" for boxes that intersect by one side
        box_2 = Rectangle(x1=0.3, y1=0.3, x2=0.5, y2=0.4)
        assert not intersection_box(box_1, box_2)
        # Checking list returned by "intersection_box" when one of boxes completely overlaps other
        box_2 = Rectangle(x1=0.1, y1=0.1, x2=0.2, y2=0.3)
        assert intersection_box(box_1, box_2) == [0.1, 0.1, 0.2, 0.3]
        # Checking list returned by "intersection_box" for boxes that not intersect
        box_2 = Rectangle(x1=0.4, y1=0.4, x2=0.6, y2=0.7)
        assert not intersection_box(box_1, box_2)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_intersection_over_union(self):
        """
        <b>Description:</b>
        Check "intersection_over_union" function

        <b>Input data:</b>
        "box_1" and "box_2" Rectangle-class objects

        <b>Expected results:</b>
        Test passes if value returned by "intersection_over_union" function is equal to expected

        <b>Steps</b>
        1. Check value returned by "intersection_over_union" function for boxes that intersect in two points
        2. Check value returned by "intersection_over_union" function for boxes that intersect in one point
        3. Check value returned by "intersection_over_union" function for boxes that intersect by one side
        4. Check value returned by "intersection_over_union" function when one of boxes completely overlaps other
        5. Check value returned by "intersection_over_union" function for boxes that not intersect
        """
        box_1 = Rectangle(x1=0.1, y1=0.1, x2=0.3, y2=0.4)
        # Checking value returned by "intersection_over_union" for boxes that intersect in two points
        box_2 = Rectangle(x1=0.2, y1=0.2, x2=0.5, y2=0.5)
        assert round(intersection_over_union(box_1, box_2), 6) == 0.153846
        # Checking value returned by "intersection_over_union" for boxes that intersect in one point
        box_2 = Rectangle(x1=0.3, y1=0.4, x2=0.5, y2=0.5)
        assert intersection_over_union(box_1, box_2) == 0.0
        # Checking value returned by "intersection_over_union" for boxes that intersect by one side
        box_2 = Rectangle(x1=0.3, y1=0.3, x2=0.5, y2=0.4)
        assert intersection_over_union(box_1, box_2) == 0.0
        # Checking value returned by "intersection_over_union" when one of boxes completely overlaps other
        box_2 = Rectangle(x1=0.1, y1=0.1, x2=0.2, y2=0.3)
        assert round(intersection_over_union(box_1, box_2), 6) == 0.333333
        # Checking value returned by "intersection_over_union" for boxes that not intersect
        box_2 = Rectangle(x1=0.4, y1=0.4, x2=0.6, y2=0.7)
        assert intersection_over_union(box_1, box_2) == 0.0

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_precision_per_class(self):
        """
        <b>Description:</b>
        Check "precision_per_class" function

        <b>Input data:</b>
        Confusion matrix

        <b>Expected results:</b>
        Test passes if array returned by "precision_per_class" function is equal to expected

        <b>Steps</b>
        1. Check array returned by "precision_per_class" function for square matrix
        2. Check array returned by "precision_per_class" function for non-square matrix
        """
        # Checking array returned by "precision_per_class" for square matrix
        matrix = np.array([(0.5, 1.0, 1.0), (1.0, 0.5, 0.5), (0.5, 1.0, 1.0)])
        assert np.array_equal(precision_per_class(matrix), np.array([0.25, 0.2, 0.4]))
        # Checking array returned by "precision_per_class" for non-square matrix
        matrix = np.array([(0.6, 0.3, 0.6), (0.3, 0.6, 0.3), (0.6, 0.8, 0.3), (0.9, 0.3, 0.6)])
        assert np.array_equal(precision_per_class(matrix), np.array([0.25, 0.3]))

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_recall_per_class(self):
        """
        <b>Description:</b>
        Check "recall_per_class" function

        <b>Input data:</b>
        Confusion matrix

        <b>Expected results:</b>
        Test passes if array returned by "recall_per_class" function is equal to expected
        """
        matrix = np.array([(6, 2, 0), (0, 10, 6), (0, 8, 12)])
        assert np.array_equal(recall_per_class(matrix), np.array([0.75, 0.625, 0.6]))

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_divide_arrays_with_possible_zeros(self):
        """
        <b>Description:</b>
        Check "divide_arrays_with_possible_zeros" function

        <b>Input data:</b>
        "numerator" and "denominator" matrices

        <b>Expected results:</b>
        Test passes if array returned by "divide_arrays_with_possible_zeros" function is equal to expected
        """
        array = np.array([(6, 2, 0), (0, 10, 6), (0, 8, 12)])
        other_array = np.array([(2, 4, 0), (0, 2, 4), (0, 0, 1)])
        assert np.array_equal(
            a1=divide_arrays_with_possible_zeros(array, other_array),
            a2=np.array([(3, 0.5, 0), (0, 5, 1.5), (0, 0, 12)]),
        )



"""Task of OTX Classification."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import json
import os
from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np
import torch

from otx.algorithms.classification.configs.base import ClassificationConfig
from otx.algorithms.classification.utils import (
    get_cls_deploy_config,
    get_cls_inferencer_configuration,
    get_cls_model_api_configuration,
)
from otx.algorithms.classification.utils import (
    get_multihead_class_info as get_hierarchical_info,
)
from otx.algorithms.common.configs import TrainType
from otx.algorithms.common.tasks.base_task import TRAIN_TYPE_DIR_PATH, OTXTask
from otx.algorithms.common.utils import embed_ir_model_data
from otx.algorithms.common.utils.callback import TrainingProgressCallback
from otx.algorithms.common.utils.logger import get_logger
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import ids_to_strings
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.explain_parameters import ExplainParameters
from otx.api.entities.inference_parameters import (
    InferenceParameters,
)
from otx.api.entities.inference_parameters import (
    default_progress_callback as default_infer_progress_callback,
)
from otx.api.entities.metadata import FloatMetadata, FloatType
from otx.api.entities.metrics import (
    CurveMetric,
    LineChartInfo,
    LineMetricsGroup,
    MetricsGroup,
    Performance,
    ScoreMetric,
)
from otx.api.entities.model import (  # ModelStatus
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
)
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.entities.train_parameters import (
    TrainParameters,
)
from otx.api.entities.train_parameters import (
    default_progress_callback as default_train_progress_callback,
)
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.utils.dataset_utils import add_saliency_maps_to_dataset_item
from otx.api.utils.labels_utils import get_empty_label
from otx.cli.utils.multi_gpu import is_multigpu_child_process

logger = get_logger()
RECIPE_TRAIN_TYPE = {
    TrainType.Semisupervised: "semisl.yaml",
    TrainType.Incremental: "incremental.yaml",
    TrainType.Selfsupervised: "selfsl.yaml",
}


class OTXClassificationTask(OTXTask, ABC):
    """Task class for OTX classification."""

    # pylint: disable=too-many-instance-attributes, too-many-locals, too-many-boolean-expressions
    def __init__(self, task_environment: TaskEnvironment, output_path: Optional[str] = None):
        super().__init__(task_environment, output_path)
        self._task_config = ClassificationConfig
        self._hyperparams = self._task_environment.get_hyper_parameters(self._task_config)
        if len(self._task_environment.get_labels(False)) == 1:
            self._labels = self._task_environment.get_labels(include_empty=True)
        else:
            self._labels = self._task_environment.get_labels(include_empty=False)
        self._empty_label = get_empty_label(self._task_environment.label_schema)

        self._multilabel = False
        self._hierarchical = False
        self._hierarchical_info = None
        self._selfsl = False
        self._set_train_mode()

        self._train_type = self._hyperparams.algo_backend.train_type
        self._model_dir = os.path.join(
            os.path.abspath(os.path.dirname(self._task_environment.model_template.model_template_path)),
            TRAIN_TYPE_DIR_PATH[self._train_type.name],
        )
        if (
            self._train_type in RECIPE_TRAIN_TYPE
            and self._train_type == TrainType.Incremental
            and not self._multilabel
            and not self._hierarchical
            and self._hyperparams.learning_parameters.enable_supcon
            and not self._model_dir.endswith("supcon")
        ):
            self._model_dir = os.path.join(self._model_dir, "supcon")

        self.data_pipeline_path = os.path.join(self._model_dir, "data_pipeline.py")

        if self._task_environment.model is not None:
            self._load_model()

    def _set_train_mode(self):
        self._multilabel = len(self._task_environment.label_schema.get_groups(False)) > 1 and len(
            self._task_environment.label_schema.get_groups(False)
        ) == len(
            self._task_environment.get_labels(include_empty=False)
        )  # noqa:E127
        if self._multilabel:
            logger.info("Classification mode: multilabel")

        if not self._multilabel and len(self._task_environment.label_schema.get_groups(False)) > 1:
            logger.info("Classification mode: hierarchical")
            self._hierarchical = True
            self._hierarchical_info = get_hierarchical_info(self._task_environment.label_schema)
        if not self._multilabel and not self._hierarchical:
            logger.info("Classification mode: multiclass")

        if self._hyperparams.algo_backend.train_type == TrainType.Selfsupervised:
            self._selfsl = True

    def infer(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Main infer function of OTX Classification."""

        logger.info("infer()")

        results = self._infer_model(dataset, inference_parameters)
        prediction_results = zip(
            results["eval_predictions"],
            results["feature_vectors"],
            results["saliency_maps"],
        )

        update_progress_callback = default_infer_progress_callback
        process_saliency_maps = False
        explain_predicted_classes = True
        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress  # type: ignore
            process_saliency_maps = inference_parameters.process_saliency_maps
            explain_predicted_classes = inference_parameters.explain_predicted_classes

        self._add_predictions_to_dataset(
            prediction_results, dataset, update_progress_callback, process_saliency_maps, explain_predicted_classes
        )
        return dataset

    def train(
        self, dataset: DatasetEntity, output_model: ModelEntity, train_parameters: Optional[TrainParameters] = None
    ):
        """Train function for OTX classification task.

        Actual training is processed by _train_model fucntion
        """
        logger.info("train()")
        # Check for stop signal when training has stopped.
        # If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # Set OTX LoggerHook & Time Monitor
        if train_parameters:
            update_progress_callback = train_parameters.update_progress
        else:
            update_progress_callback = default_train_progress_callback
        self._time_monitor = TrainingProgressCallback(update_progress_callback)

        results = self._train_model(dataset)

        # Check for stop signal when training has stopped. If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # get output model
        model_ckpt = results.get("final_ckpt")
        if model_ckpt is None:
            logger.error("cannot find final checkpoint from the results.")
            return
        # update checkpoint to the newly trained model
        self._model_ckpt = model_ckpt

        # compose performance statistics
        training_metrics, final_acc = self._generate_training_metrics(self._learning_curves)
        # save resulting model
        self.save_model(output_model)
        performance = Performance(
            score=ScoreMetric(value=final_acc, name="accuracy"),
            dashboard_metrics=training_metrics,
        )
        logger.info(f"Final model performance: {str(performance)}")
        output_model.performance = performance
        self._is_training = False
        logger.info("train done.")

    def export(
        self,
        export_type: ExportType,
        output_model: ModelEntity,
        precision: ModelPrecision = ModelPrecision.FP32,
        dump_features: bool = True,
    ):
        """Export function of OTX Classification Task."""

        logger.info("Exporting the model")
        if export_type != ExportType.OPENVINO:
            raise RuntimeError(f"not supported export type {export_type}")
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.MO

        results = self._export_model(precision, dump_features)
        outputs = results.get("outputs")
        logger.debug(f"results of run_task = {outputs}")
        if outputs is None:
            raise RuntimeError(results.get("msg"))

        bin_file = outputs.get("bin")
        xml_file = outputs.get("xml")
        onnx_file = outputs.get("onnx")

        inference_config = get_cls_inferencer_configuration(self._task_environment.label_schema)
        deploy_cfg = get_cls_deploy_config(self._task_environment.label_schema, inference_config)
        ir_extra_data = get_cls_model_api_configuration(self._task_environment.label_schema, inference_config)
        ir_extra_data[("otx_config",)] = json.dumps(deploy_cfg, ensure_ascii=False)
        embed_ir_model_data(xml_file, ir_extra_data)

        if xml_file is None or bin_file is None or onnx_file is None:
            raise RuntimeError("invalid status of exporting. bin and xml or onnx should not be None")
        with open(bin_file, "rb") as f:
            output_model.set_data("openvino.bin", f.read())
        with open(xml_file, "rb") as f:
            output_model.set_data("openvino.xml", f.read())
        with open(onnx_file, "rb") as f:
            output_model.set_data("model.onnx", f.read())
        output_model.precision = self._precision
        output_model.has_xai = dump_features
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        logger.info("Exporting completed")

    def explain(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[ExplainParameters] = None,
    ) -> DatasetEntity:
        """Main explain function of OTX Classification Task."""

        predictions, saliency_maps = self._explain_model(
            dataset,
            explain_parameters=explain_parameters,
        )

        update_progress_callback = default_infer_progress_callback
        process_saliency_maps = False
        explain_predicted_classes = True
        if explain_parameters is not None:
            update_progress_callback = explain_parameters.update_progress  # type: ignore
            process_saliency_maps = explain_parameters.process_saliency_maps
            explain_predicted_classes = explain_parameters.explain_predicted_classes

        self._add_explanations_to_dataset(
            predictions,
            saliency_maps,
            dataset,
            update_progress_callback,
            process_saliency_maps,
            explain_predicted_classes,
        )
        logger.info("Explain completed")
        return dataset

    def evaluate(
        self,
        output_resultset: ResultSetEntity,
        evaluation_metric: Optional[str] = None,
    ):
        """Evaluate function of OTX Classification Task."""

        logger.info("called evaluate()")
        if evaluation_metric is not None:
            logger.warning(
                f"Requested to use {evaluation_metric} metric, " "but parameter is ignored. Use accuracy instead."
            )
        metric = MetricsHelper.compute_accuracy(output_resultset)
        logger.info(f"Accuracy after evaluation: {metric.accuracy.value}")
        output_resultset.performance = metric.get_performance()
        logger.info("Evaluation completed")

    # pylint: disable=too-many-branches, too-many-locals
    def _add_predictions_to_dataset(
        self,
        prediction_results,
        dataset,
        update_progress_callback,
        process_saliency_maps=False,
        explain_predicted_classes=True,
    ):
        """Loop over dataset again to assign predictions.Convert from MMClassification format to OTX format."""

        dataset_size = len(dataset)
        pos_thr = 0.5
        for i, (dataset_item, prediction_items) in enumerate(zip(dataset, prediction_results)):
            prediction_item, feature_vector, saliency_map = prediction_items
            if any(np.isnan(prediction_item)):
                logger.info("Nan in prediction_item.")

            item_labels = self._get_item_labels(prediction_item, pos_thr)

            dataset_item.append_labels(item_labels)

            probs = TensorEntity(name="probabilities", numpy=prediction_item.reshape(-1))
            dataset_item.append_metadata_item(probs, model=self._task_environment.model)

            top_idxs = np.argpartition(prediction_item, -2)[-2:]
            top_probs = prediction_item[top_idxs]
            active_score_media = FloatMetadata(
                name="active_score", value=top_probs[1] - top_probs[0], float_type=FloatType.ACTIVE_SCORE
            )
            dataset_item.append_metadata_item(active_score_media, model=self._task_environment.model)

            if feature_vector is not None:
                active_score = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(active_score, model=self._task_environment.model)

            if saliency_map is not None:
                add_saliency_maps_to_dataset_item(
                    dataset_item=dataset_item,
                    saliency_map=saliency_map,
                    model=self._task_environment.model,
                    labels=self._labels,
                    predicted_scored_labels=item_labels,
                    explain_predicted_classes=explain_predicted_classes,
                    process_saliency_maps=process_saliency_maps,
                )
            update_progress_callback(int(i / dataset_size * 100))

    # pylint: disable=too-many-locals
    def _get_item_labels(self, prediction_item, pos_thr):
        item_labels = []

        if self._multilabel:
            if max(prediction_item) < pos_thr:
                logger.info("Confidence is smaller than pos_thr, empty_label will be appended to item_labels.")
                item_labels.append(ScoredLabel(self._empty_label, probability=1.0))
            else:
                for cls_idx, pred_item in enumerate(prediction_item):
                    if pred_item > pos_thr:
                        cls_label = ScoredLabel(self._labels[cls_idx], probability=float(pred_item))
                        item_labels.append(cls_label)

        elif self._hierarchical:
            for head_idx in range(self._hierarchical_info["num_multiclass_heads"]):
                logits_begin, logits_end = self._hierarchical_info["head_idx_to_logits_range"][str(head_idx)]
                head_logits = prediction_item[logits_begin:logits_end]
                head_pred = np.argmax(head_logits)  # Assume logits already passed softmax
                label_str = self._hierarchical_info["all_groups"][head_idx][head_pred]
                otx_label = next(x for x in self._labels if x.name == label_str)
                item_labels.append(ScoredLabel(label=otx_label, probability=float(head_logits[head_pred])))

            if self._hierarchical_info["num_multilabel_classes"]:
                head_logits = prediction_item[self._hierarchical_info["num_single_label_classes"] :]
                for logit_idx, logit in enumerate(head_logits):
                    if logit > pos_thr:  # Assume logits already passed sigmoid
                        label_str_idx = self._hierarchical_info["num_multiclass_heads"] + logit_idx
                        label_str = self._hierarchical_info["all_groups"][label_str_idx][0]
                        otx_label = next(x for x in self._labels if x.name == label_str)
                        item_labels.append(ScoredLabel(label=otx_label, probability=float(logit)))
            item_labels = self._task_environment.label_schema.resolve_labels_probabilistic(item_labels)
            if not item_labels:
                logger.info("item_labels is empty.")
                item_labels.append(ScoredLabel(self._empty_label, probability=1.0))

        else:
            label_idx = prediction_item.argmax()
            cls_label = ScoredLabel(
                self._labels[label_idx],
                probability=float(prediction_item[label_idx]),
            )
            item_labels.append(cls_label)
        return item_labels

    def _add_explanations_to_dataset(
        self,
        predictions,
        saliency_maps,
        dataset,
        update_progress_callback,
        process_saliency_maps,
        explain_predicted_classes,
    ):
        """Loop over dataset again and assign saliency maps."""
        dataset_size = len(dataset)
        for i, (dataset_item, prediction_item, saliency_map) in enumerate(zip(dataset, predictions, saliency_maps)):
            item_labels = self._get_item_labels(prediction_item, pos_thr=0.5)
            add_saliency_maps_to_dataset_item(
                dataset_item=dataset_item,
                saliency_map=saliency_map,
                model=self._task_environment.model,
                labels=self._labels,
                predicted_scored_labels=item_labels,
                explain_predicted_classes=explain_predicted_classes,
                process_saliency_maps=process_saliency_maps,
            )
            update_progress_callback(int(i / dataset_size * 100))

    def save_model(self, output_model: ModelEntity):
        """Save best model weights in ClassificationTrainTask."""
        if is_multigpu_child_process():
            return

        logger.info("called save_model")
        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}
        model_ckpt = torch.load(self._model_ckpt)
        modelinfo = {
            "model": model_ckpt,
            "config": hyperparams_str,
            "labels": labels,
            "VERSION": 1,
        }

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        output_model.precision = self._precision

    def _generate_training_metrics(self, learning_curves):  # pylint: disable=arguments-renamed
        """Parses the classification logs to get metrics from the latest training run.

        :return output List[MetricsGroup]
        """
        output: List[MetricsGroup] = []

        if self._multilabel:
            metric_key = "val/accuracy-mlc"
        elif self._hierarchical:
            metric_key = "val/MHAcc"
        else:
            metric_key = "val/accuracy_top-1"

        # Learning curves
        best_acc = -1
        if learning_curves is None:
            return output

        for key, curve in learning_curves.items():
            metric_curve = CurveMetric(xs=curve.x, ys=curve.y, name=key)
            if key == metric_key:
                best_acc = max(curve.y)
            visualization_info = LineChartInfo(name=key, x_axis_label="Timestamp", y_axis_label=key)
            output.append(LineMetricsGroup(metrics=[metric_curve], visualization_info=visualization_info))

        return output, best_acc

    @abstractmethod
    def _train_model(self, dataset: DatasetEntity):
        """Train model and return the results."""
        raise NotImplementedError

    @abstractmethod
    def _export_model(self, precision, dump_features):
        """Export model and return the results."""
        raise NotImplementedError

    @abstractmethod
    def _explain_model(self, dataset: DatasetEntity, explain_parameters: Optional[ExplainParameters]):
        """Explain model and return the results."""
        raise NotImplementedError

    @abstractmethod
    def _infer_model(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ):
        """Get inference results from dataset."""
        raise NotImplementedError



"""Builder module for otx.core.ov.graph.parsers."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from otx.core.ov.registry import Registry

PARSERS = Registry("ov graph parsers")



"""OTX Adapters for action recognition backbones - mmaction2."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .movinet_recognizer import MoViNetRecognizer

__all__ = ["MoViNetRecognizer"]



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
from unittest.mock import patch

import pytest

from otx.api.entities.interfaces.graph_interface import IGraph
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIGraph:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @patch("otx.api.entities.interfaces.graph_interface.IGraph.__abstractmethods__", set())
    def test_i_graph(self):
        """
        <b>Description:</b>
        Check IGraph class object initialization

        <b>Input data:</b>
        IGraph object

        <b>Expected results:</b>
        Test passes if IGraph object methods raise NotImplementedError exception
        """
        i_graph = IGraph()
        with pytest.raises(NotImplementedError):
            i_graph.add_node(1)
        with pytest.raises(NotImplementedError):
            i_graph.add_edge(1, 2)
        with pytest.raises(NotImplementedError):
            i_graph.has_edge_between(1, 2)
        with pytest.raises(NotImplementedError):
            i_graph.neighbors(1)
        with pytest.raises(NotImplementedError):
            i_graph.find_cliques()
        with pytest.raises(NotImplementedError):
            i_graph.num_nodes()
        with pytest.raises(NotImplementedError):
            i_graph.remove_edges(1, 2)
        with pytest.raises(NotImplementedError):
            i_graph.find_out_edges(1)
        with pytest.raises(NotImplementedError):
            i_graph.find_in_edges(1)
        with pytest.raises(NotImplementedError):
            i_graph.edges
        with pytest.raises(NotImplementedError):
            i_graph.nodes



"""Unit Test for otx.algorithms.action.adapters.mmaction.heads.movinet_head."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
import torch

from otx.algorithms.action.adapters.mmaction.models.heads.movinet_head import (
    MoViNetHead,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestMoViNetHead:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.movinet_head = MoViNetHead(
            num_classes=400,
            in_channels=480,
            hidden_dim=2048,
            loss_cls=dict(type="CrossEntropyLoss", loss_weight=1.0),
        )

    @e2e_pytest_unit
    def test_forward(self) -> None:
        """Test forward function."""
        sample_input = torch.randn(1, 480, 1, 1, 1)
        with torch.no_grad():
            out = self.movinet_head(sample_input)
        assert out.shape == (1, self.movinet_head.num_classes)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import pytest

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label_schema import LabelGroup, LabelSchemaEntity
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.usecases.reporting.callback import Callback
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestCallback:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_callback_attributes(self):
        """
        <b>Description:</b>
        Check "Callback" class object initialization

        <b>Input data:</b>
        "Callback" class object

        <b>Expected results:</b>
        Test passes if "params" and "model" attributes of initialized "Callback" class object are equal to expected

        <b>Steps</b>
        1. Check "model" attribute of "Callback" object after "set_model" method
        2. Check "params" attribute of "Callback" object after "set_params" method
        """
        callback = Callback()
        # Checking "params" of "Callback" object after "set_params"
        params = {"parameter_1": 1, "parameter_2": 4, "parameter_3": 9}
        callback.set_params(params)
        assert callback.params == params
        # Checking "model" of "Callback" after "set_model"
        configurable_params = ConfigurableParameters(header="Test model configurable params")
        labels_group = LabelGroup(name="model_group", labels=[])
        model_configuration = ModelConfiguration(configurable_params, LabelSchemaEntity(label_groups=[labels_group]))
        model = ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)
        callback.set_model(model)
        assert callback.model == model

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_callback_abstract_methods(self):
        """
        <b>Description:</b>
        Check "Callback" class object abstract methods

        <b>Input data:</b>
        "Callback" class object

        <b>Expected results:</b>
        Test passes if none of "Callback" class abstract methods raised an exception
        """
        callback = Callback()
        callback.on_epoch_begin(epoch=1, logs="on_epoch_begin logs")
        callback.on_epoch_end(epoch=1, logs="on_epoch_end logs")
        callback.on_batch_begin(batch=1, logs="on_batch_begin logs")
        callback.on_batch_end(batch=1, logs="on_batch_end logs")
        callback.on_train_begin(logs="on_train_begin logs")
        callback.on_train_end(logs="on_train_end logs")
        callback.on_train_batch_begin(batch=1, logs="on_train_batch_begin logs")
        callback.on_train_batch_end(batch=1, logs="on_train_batch_end logs")
        callback.on_test_begin("on_test_begin logs")
        callback.on_test_end("on_epoch_begin logs")
        callback.on_test_batch_begin(batch=1, logs="on_test_batch_begin logs")
        callback.on_test_batch_end(batch=1, logs="on_epoch_begin logs")



"""OTX ATSS Class for mmdetection detectors."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import functools

import torch
from mmdet.models.builder import DETECTORS
from mmdet.models.detectors.atss import ATSS

from otx.algorithms.common.adapters.mmcv.hooks.recording_forward_hook import (
    FeatureVectorHook,
)
from otx.algorithms.common.adapters.mmdeploy.utils import is_mmdeploy_enabled
from otx.algorithms.common.utils.logger import get_logger
from otx.algorithms.common.utils.task_adapt import map_class_names
from otx.algorithms.detection.adapters.mmdet.hooks.det_class_probability_map_hook import (
    DetClassProbabilityMapHook,
)

from .l2sp_detector_mixin import L2SPDetectorMixin
from .sam_detector_mixin import SAMDetectorMixin

logger = get_logger()

# TODO: Need to fix pylint issues
# pylint: disable=abstract-method, unused-argument, too-many-locals, protected-access


@DETECTORS.register_module()
class CustomATSS(SAMDetectorMixin, L2SPDetectorMixin, ATSS):
    """SAM optimizer & L2SP regularizer enabled custom ATSS."""

    def __init__(self, *args, task_adapt=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Hook for class-sensitive weight loading
        if task_adapt:
            self._register_load_state_dict_pre_hook(
                functools.partial(
                    self.load_state_dict_pre_hook,
                    self,  # model
                    task_adapt["dst_classes"],  # model_classes
                    task_adapt["src_classes"],  # chkpt_classes
                )
            )

    def forward_train(self, img, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore=None, **kwargs):
        """Forward function for CustomATSS."""
        return super().forward_train(img, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore=gt_bboxes_ignore)

    @staticmethod
    def load_state_dict_pre_hook(model, model_classes, chkpt_classes, chkpt_dict, prefix, *args, **kwargs):
        """Modify input state_dict according to class name matching before weight loading."""
        logger.info(f"----------------- CustomATSS.load_state_dict_pre_hook() called w/ prefix: {prefix}")

        # Dst to src mapping index
        model_classes = list(model_classes)
        chkpt_classes = list(chkpt_classes)
        model2chkpt = map_class_names(model_classes, chkpt_classes)
        logger.info(f"{chkpt_classes} -> {model_classes} ({model2chkpt})")

        model_dict = model.state_dict()
        param_names = [
            "bbox_head.atss_cls.weight",
            "bbox_head.atss_cls.bias",
        ]
        for model_name in param_names:
            chkpt_name = prefix + model_name
            if model_name not in model_dict or chkpt_name not in chkpt_dict:
                logger.info(f"Skipping weight copy: {chkpt_name}")
                continue

            # Mix weights
            model_param = model_dict[model_name].clone()
            chkpt_param = chkpt_dict[chkpt_name]
            for model_t, ckpt_t in enumerate(model2chkpt):
                if ckpt_t >= 0:
                    model_param[model_t].copy_(chkpt_param[ckpt_t])

            # Replace checkpoint weight by mixed weights
            chkpt_dict[chkpt_name] = model_param


if is_mmdeploy_enabled():
    from mmdeploy.core import FUNCTION_REWRITER, mark
    from mmdeploy.utils import is_dynamic_shape

    @FUNCTION_REWRITER.register_rewriter(
        "otx.algorithms.detection.adapters.mmdet.models.detectors.custom_atss_detector.CustomATSS.simple_test"
    )
    def custom_atss__simple_test(ctx, self, img, img_metas, **kwargs):
        """Function for custom_atss__simple_test."""
        feat = self.extract_feat(img)
        outs = self.bbox_head(feat)
        bbox_results = self.bbox_head.get_bboxes(*outs, img_metas=img_metas, cfg=self.test_cfg, **kwargs)

        if ctx.cfg["dump_features"]:
            feature_vector = FeatureVectorHook.func(feat)
            cls_scores = outs[0]
            saliency_map = DetClassProbabilityMapHook(self).func(feature_map=cls_scores, cls_scores_provided=True)
            return (*bbox_results, feature_vector, saliency_map)

        return bbox_results

    @mark("custom_atss_forward", inputs=["input"], outputs=["dets", "labels", "feats", "saliencies"])
    def __forward_impl(ctx, self, img, img_metas, **kwargs):
        """Internal Function for __forward_impl."""
        assert isinstance(img, torch.Tensor)

        deploy_cfg = ctx.cfg
        is_dynamic_flag = is_dynamic_shape(deploy_cfg)
        # get origin input shape as tensor to support onnx dynamic shape
        img_shape = torch._shape_as_tensor(img)[2:]
        if not is_dynamic_flag:
            img_shape = [int(val) for val in img_shape]
        img_metas[0]["img_shape"] = img_shape
        return self.simple_test(img, img_metas, **kwargs)

    @FUNCTION_REWRITER.register_rewriter(
        "otx.algorithms.detection.adapters.mmdet.models.detectors.custom_atss_detector.CustomATSS.forward"
    )
    def custom_atss__forward(ctx, self, img, img_metas=None, return_loss=False, **kwargs):
        """Internal Function for __forward for CustomATSS."""
        if img_metas is None:
            img_metas = [{}]
        else:
            assert len(img_metas) == 1, "do not support aug_test"
            img_metas = img_metas[0]

        if isinstance(img, list):
            img = img[0]

        return __forward_impl(ctx, self, img, img_metas=img_metas, **kwargs)



"""Task of OTX Video Recognition."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
import torch
from mmcv.utils import ConfigDict

from otx.algorithms.action.configs.base import ActionConfig
from otx.algorithms.common.tasks.base_task import TRAIN_TYPE_DIR_PATH, OTXTask
from otx.algorithms.common.utils.callback import (
    InferenceProgressCallback,
    TrainingProgressCallback,
)
from otx.algorithms.common.utils.logger import get_logger
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import config_to_bytes, ids_to_strings
from otx.api.entities.annotation import Annotation
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.explain_parameters import ExplainParameters
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.metrics import (
    BarChartInfo,
    BarMetricsGroup,
    CurveMetric,
    LineChartInfo,
    LineMetricsGroup,
    MetricsGroup,
    ScoreMetric,
    VisualizationType,
)
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
)
from otx.api.entities.model_template import TaskType
from otx.api.entities.result_media import ResultMediaEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.entities.train_parameters import TrainParameters, default_progress_callback
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.evaluation.accuracy import Accuracy
from otx.api.usecases.evaluation.f_measure import FMeasure
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.utils.vis_utils import get_actmap

logger = get_logger()


class OTXActionTask(OTXTask, ABC):
    """Task class for OTX action."""

    # pylint: disable=too-many-instance-attributes, too-many-locals
    def __init__(self, task_environment: TaskEnvironment, output_path: Optional[str] = None):
        super().__init__(task_environment, output_path)
        self._task_config = ActionConfig
        self._hyperparams: ConfigDict = task_environment.get_hyper_parameters(self._task_config)
        self._train_type = self._hyperparams.algo_backend.train_type
        self._model_dir = os.path.join(
            os.path.abspath(os.path.dirname(self._task_environment.model_template.model_template_path)),
            TRAIN_TYPE_DIR_PATH[self._train_type.name],
        )

        if hasattr(self._hyperparams, "postprocessing") and hasattr(
            self._hyperparams.postprocessing, "confidence_threshold"
        ):
            self.confidence_threshold = self._hyperparams.postprocessing.confidence_threshold
        else:
            self.confidence_threshold = 0.0

        if task_environment.model is not None:
            self._load_model()

        self.data_pipeline_path = os.path.join(self._model_dir, "data_pipeline.py")

    def train(
        self, dataset: DatasetEntity, output_model: ModelEntity, train_parameters: Optional[TrainParameters] = None
    ):
        """Train function for OTX action task.

        Actual training is processed by _train_model fucntion
        """
        logger.info("train()")
        # Check for stop signal when training has stopped.
        # If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # Set OTX LoggerHook & Time Monitor
        if train_parameters:
            update_progress_callback = train_parameters.update_progress
        else:
            update_progress_callback = default_progress_callback
        self._time_monitor = TrainingProgressCallback(update_progress_callback)

        results = self._train_model(dataset)

        # Check for stop signal when training has stopped. If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # get output model
        model_ckpt = results.get("final_ckpt")
        if model_ckpt is None:
            logger.error("cannot find final checkpoint from the results.")
            return
        # update checkpoint to the newly trained model
        self._model_ckpt = model_ckpt

        # get prediction on validation set
        self._is_training = False
        val_dataset = dataset.get_subset(Subset.VALIDATION)
        val_preds, val_performance = self._infer_model(val_dataset, InferenceParameters(is_evaluation=True))

        preds_val_dataset = val_dataset.with_empty_annotations()
        if self._task_type == TaskType.ACTION_CLASSIFICATION:
            self._add_cls_predictions_to_dataset(val_preds, preds_val_dataset)
        elif self._task_type == TaskType.ACTION_DETECTION:
            self._add_det_predictions_to_dataset(val_preds, preds_val_dataset, 0.0)

        result_set = ResultSetEntity(
            model=output_model,
            ground_truth_dataset=val_dataset,
            prediction_dataset=preds_val_dataset,
        )

        metric: Union[Accuracy, FMeasure]

        if self._task_type == TaskType.ACTION_CLASSIFICATION:
            metric = MetricsHelper.compute_accuracy(result_set)
        if self._task_type == TaskType.ACTION_DETECTION:
            if self._hyperparams.postprocessing.result_based_confidence_threshold:
                best_confidence_threshold = None
                logger.info("Adjusting the confidence threshold")
                metric = MetricsHelper.compute_f_measure(result_set, vary_confidence_threshold=True)
                if metric.best_confidence_threshold:
                    best_confidence_threshold = metric.best_confidence_threshold.value
                if best_confidence_threshold is None:
                    raise ValueError("Cannot compute metrics: Invalid confidence threshold!")
                logger.info(f"Setting confidence threshold to {best_confidence_threshold} based on results")
                self.confidence_threshold = best_confidence_threshold
            else:
                metric = MetricsHelper.compute_f_measure(result_set, vary_confidence_threshold=False)

        # compose performance statistics
        performance = metric.get_performance()
        performance.dashboard_metrics.extend(self._generate_training_metrics(self._learning_curves, val_performance))
        logger.info(f"Final model performance: {performance}")
        # save resulting model
        self.save_model(output_model)
        output_model.performance = performance
        logger.info("train done.")

    @abstractmethod
    def _train_model(self, dataset: DatasetEntity):
        """Train model and return the results."""
        raise NotImplementedError

    def infer(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Main infer function."""
        logger.info("infer()")

        update_progress_callback = default_progress_callback
        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress  # type: ignore

        self._time_monitor = InferenceProgressCallback(len(dataset), update_progress_callback)
        # If confidence threshold is adaptive then up-to-date value should be stored in the model
        # and should not be changed during inference. Otherwise user-specified value should be taken.
        if not self._hyperparams.postprocessing.result_based_confidence_threshold:
            self.confidence_threshold = self._hyperparams.postprocessing.confidence_threshold
        logger.info(f"Confidence threshold {self.confidence_threshold}")

        prediction_results, _ = self._infer_model(dataset, inference_parameters)

        if self._task_type == TaskType.ACTION_CLASSIFICATION:
            self._add_cls_predictions_to_dataset(prediction_results, dataset)
        elif self._task_type == TaskType.ACTION_DETECTION:
            self._add_det_predictions_to_dataset(prediction_results, dataset, self.confidence_threshold)
        logger.info("Inference completed")
        return dataset

    @abstractmethod
    def _infer_model(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ):
        """Get inference results from dataset."""
        raise NotImplementedError

    def export(
        self,
        export_type: ExportType,
        output_model: ModelEntity,
        precision: ModelPrecision = ModelPrecision.FP32,
        dump_features: bool = True,
    ):
        """Export function of OTX Task."""
        if dump_features:
            raise NotImplementedError(
                "Feature dumping is not implemented for the action task."
                "The saliency maps and representation vector outputs will not be dumped in the exported model."
            )

        # copied from OTX inference_task.py
        logger.info("Exporting the model")
        if export_type != ExportType.OPENVINO:
            raise RuntimeError(f"not supported export type {export_type}")
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.MO
        output_model.has_xai = dump_features

        results = self._export_model(precision, dump_features)

        outputs = results.get("outputs")
        logger.debug(f"results of run_task = {outputs}")
        if outputs is None:
            raise RuntimeError(results.get("msg"))

        bin_file = outputs.get("bin")
        xml_file = outputs.get("xml")
        onnx_file = outputs.get("onnx")

        if xml_file is None or bin_file is None or onnx_file is None:
            raise RuntimeError("invalid status of exporting. bin and xml should not be None")
        with open(bin_file, "rb") as f:
            output_model.set_data("openvino.bin", f.read())
        with open(xml_file, "rb") as f:
            output_model.set_data("openvino.xml", f.read())
        with open(onnx_file, "rb") as f:
            output_model.set_data("model.onnx", f.read())
        output_model.set_data(
            "confidence_threshold",
            np.array([self.confidence_threshold], dtype=np.float32).tobytes(),
        )
        output_model.set_data("config.json", config_to_bytes(self._hyperparams))
        output_model.precision = self._precision
        output_model.optimization_methods = self._optimization_methods
        output_model.has_xai = dump_features
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        logger.info("Exporting completed")

    @abstractmethod
    def _export_model(self, precision: ModelPrecision, dump_features: bool = True):
        raise NotImplementedError

    def explain(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[ExplainParameters] = None,
    ) -> DatasetEntity:
        """Main explain function of OTX Task."""
        raise NotImplementedError("Video recognition task don't support otx explain yet.")

    def evaluate(
        self,
        output_resultset: ResultSetEntity,
        evaluation_metric: Optional[str] = None,
    ):
        """Evaluate function of OTX Action Task."""
        logger.info("called evaluate()")
        if evaluation_metric is not None:
            logger.warning(
                f"Requested to use {evaluation_metric} metric, " "but parameter is ignored. Use F-measure instead."
            )
        self._remove_empty_frames(output_resultset.ground_truth_dataset)

        metric: Union[Accuracy, FMeasure]
        if self._task_type == TaskType.ACTION_CLASSIFICATION:
            metric = MetricsHelper.compute_accuracy(output_resultset)
        if self._task_type == TaskType.ACTION_DETECTION:
            metric = MetricsHelper.compute_f_measure(output_resultset)
        performance = metric.get_performance()
        logger.info(f"Final model performance: {str(performance)}")
        output_resultset.performance = metric.get_performance()
        logger.info("Evaluation completed")

    def _remove_empty_frames(self, dataset: DatasetEntity):
        """Remove empty frame for action detection dataset."""
        remove_indices = []
        for idx, item in enumerate(dataset):
            if item.get_metadata()[0].data.is_empty_frame:
                remove_indices.append(idx)
        dataset.remove_at_indices(remove_indices)

    def _add_cls_predictions_to_dataset(self, prediction_results: Iterable, dataset: DatasetEntity):
        """Loop over dataset again to assign predictions. Convert from MM format to OTX format."""
        prediction_results = list(prediction_results)
        video_info: Dict[str, int] = {}
        for dataset_item in dataset:
            video_id = dataset_item.get_metadata()[0].data.video_id
            if video_id not in video_info:
                video_info[video_id] = len(video_info)
        for dataset_item in dataset:
            video_id = dataset_item.get_metadata()[0].data.video_id
            all_results, feature_vector, saliency_map = prediction_results[video_info[video_id]]
            item_labels = []
            label = ScoredLabel(label=self._labels[all_results.argmax()], probability=all_results.max())
            item_labels.append(label)
            dataset_item.append_labels(item_labels)

            if feature_vector is not None:
                active_score = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(active_score, model=self._task_environment.model)

            if saliency_map is not None:
                saliency_map = get_actmap(saliency_map, (dataset_item.width, dataset_item.height))
                saliency_map_media = ResultMediaEntity(
                    name="Saliency Map",
                    type="saliency_map",
                    annotation_scene=dataset_item.annotation_scene,
                    numpy=saliency_map,
                    roi=dataset_item.roi,
                )
                dataset_item.append_metadata_item(saliency_map_media, model=self._task_environment.model)

    def _add_det_predictions_to_dataset(
        self, prediction_results: Iterable, dataset: DatasetEntity, confidence_threshold: float = 0.05
    ):
        self._remove_empty_frames(dataset)
        for dataset_item, (all_results, feature_vector, saliency_map) in zip(dataset, prediction_results):
            shapes = []
            for label_idx, detections in enumerate(all_results):
                for i in range(detections.shape[0]):
                    probability = float(detections[i, 4])
                    coords = detections[i, :4]

                    if probability < confidence_threshold:
                        continue
                    if coords[3] - coords[1] <= 0 or coords[2] - coords[0] <= 0:
                        continue

                    assigned_label = [ScoredLabel(self._labels[label_idx], probability=probability)]
                    shapes.append(
                        Annotation(
                            Rectangle(x1=coords[0], y1=coords[1], x2=coords[2], y2=coords[3]),
                            labels=assigned_label,
                        )
                    )
            dataset_item.append_annotations(shapes)

            if feature_vector is not None:
                active_score = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(active_score, model=self._task_environment.model)

            if saliency_map is not None:
                saliency_map = get_actmap(saliency_map, (dataset_item.width, dataset_item.height))
                saliency_map_media = ResultMediaEntity(
                    name="Saliency Map",
                    type="saliency_map",
                    annotation_scene=dataset_item.annotation_scene,
                    numpy=saliency_map,
                    roi=dataset_item.roi,
                )
                dataset_item.append_metadata_item(saliency_map_media, model=self._task_environment.model)

    @staticmethod
    # TODO Implement proper function for action classification
    def _generate_training_metrics(learning_curves, scores, metric_name="mAP") -> Iterable[MetricsGroup[Any, Any]]:
        """Get Training metrics (epochs & scores).

        Parses the mmaction logs to get metrics from the latest training run
        :return output List[MetricsGroup]
        """
        output: List[MetricsGroup] = []

        # Learning curves.
        for key, curve in learning_curves.items():
            len_x, len_y = len(curve.x), len(curve.y)
            if len_x != len_y:
                logger.warning(f"Learning curve {key} has inconsistent number of coordinates ({len_x} vs {len_y}.")
                len_x = min(len_x, len_y)
                curve.x = curve.x[:len_x]
                curve.y = curve.y[:len_x]
            metric_curve = CurveMetric(
                xs=np.nan_to_num(curve.x).tolist(),
                ys=np.nan_to_num(curve.y).tolist(),
                name=key,
            )
            visualization_info = LineChartInfo(name=key, x_axis_label="Epoch", y_axis_label=key)
            output.append(LineMetricsGroup(metrics=[metric_curve], visualization_info=visualization_info))

        # Final mAP value on the validation set.
        output.append(
            BarMetricsGroup(
                metrics=[ScoreMetric(value=scores, name=f"{metric_name}")],
                visualization_info=BarChartInfo("Validation score", visualization_type=VisualizationType.RADIAL_BAR),
            )
        )

        return output

    def save_model(self, output_model: ModelEntity):
        """Save best model weights in ActionTrainTask."""
        logger.info("called save_model")
        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}
        model_ckpt = torch.load(self._model_ckpt)
        modelinfo = {
            "model": model_ckpt["state_dict"],
            "config": hyperparams_str,
            "labels": labels,
            "confidence_threshold": self.confidence_threshold,
            "VERSION": 1,
        }

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        output_model.precision = self._precision



"""OTX Adapters - mmseg."""


# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .datasets import MPASegDataset
from .models import (
    ConstantScalarScheduler,
    CrossEntropyLossWithIgnore,
    CustomFCNHead,
    DetConB,
    DetConLoss,
    LiteHRNet,
    MeanTeacherSegmentor,
    MMOVBackbone,
    MMOVDecodeHead,
    PolyScalarScheduler,
    SelfSLMLP,
    StepScalarScheduler,
    SupConDetConB,
)

# fmt: off
# isort: off
# FIXME: openvino pot library adds stream handlers to root logger
# which makes annoying duplicated logging
# pylint: disable=no-name-in-module,wrong-import-order
from mmseg.utils import get_root_logger  # type: ignore # (false positive)
get_root_logger().propagate = False
# fmt: off
# isort: on

__all__ = [
    "MPASegDataset",
    "LiteHRNet",
    "MMOVBackbone",
    "CustomFCNHead",
    "MMOVDecodeHead",
    "DetConLoss",
    "SelfSLMLP",
    "ConstantScalarScheduler",
    "PolyScalarScheduler",
    "StepScalarScheduler",
    "DetConB",
    "CrossEntropyLossWithIgnore",
    "SupConDetConB",
    "MeanTeacherSegmentor",
]



"""Collection of utils for task implementation in Classification Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import List, Optional

from mmcv import Config, ConfigDict

from otx.algorithms.common.adapters.mmcv.utils import (
    get_configs_by_pairs,
    get_dataset_configs,
    get_meta_keys,
    patch_color_conversion,
)
from otx.algorithms.common.utils.logger import get_logger
from otx.api.entities.label import Domain

logger = get_logger()


def patch_datasets(
    config: Config,
    domain: Domain = Domain.CLASSIFICATION,
    subsets: Optional[List[str]] = None,
    **kwargs,
):
    """Update dataset configs."""
    assert "data" in config
    assert "type" in kwargs

    if subsets is None:
        subsets = ["train", "val", "test", "unlabeled"]

    def update_pipeline(cfg):
        if subset == "train":
            for collect_cfg in get_configs_by_pairs(cfg, dict(type="Collect")):
                get_meta_keys(collect_cfg, ["entity_id", "label_id"])

    for subset in subsets:
        if subset not in config.data:
            continue
        config.data[f"{subset}_dataloader"] = config.data.get(f"{subset}_dataloader", ConfigDict())

        # For stable hierarchical information indexing
        if subset == "train" and kwargs["type"] == "OTXHierarchicalClsDataset":
            config.data[f"{subset}_dataloader"]["drop_last"] = True

        cfgs = get_dataset_configs(config, subset)
        for cfg in cfgs:
            cfg.domain = domain
            cfg.otx_dataset = None
            cfg.labels = None
            cfg.update(kwargs)

            update_pipeline(cfg)

    patch_color_conversion(config)


def patch_evaluation(config: Config, task: str):
    """Patch evaluation."""
    cfg = config.get("evaluation", None)
    if cfg:
        if task == "multilabel":
            cfg.metric = ["accuracy-mlc", "mAP", "CP", "OP", "CR", "OR", "CF1", "OF1"]
            config.early_stop_metric = "mAP"
        elif task == "hierarchical":
            cfg.metric = ["MHAcc", "avgClsAcc", "mAP"]
            config.early_stop_metric = "MHAcc"
        elif task == "normal":
            cfg.metric = ["accuracy", "class_accuracy"]
            config.early_stop_metric = "accuracy"
        else:
            raise NotImplementedError



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from functools import partial

import pytest

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.elements import (
    ParameterGroup,
    configurable_boolean,
    configurable_float,
    configurable_integer,
    float_selectable,
    selectable,
)
from otx.api.configuration.helper.config_element_mapping import (
    GroupElementMapping,
    PrimitiveElementMapping,
    RuleElementMapping,
)
from otx.api.configuration.ui_rules.rules import Rule, UIRules
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestPrimitiveElementMapping:
    @staticmethod
    def check_primitive_element(
        primitive_element: PrimitiveElementMapping,
        expected_name: str,
        expected_partial_element: partial,
    ):
        assert primitive_element.name == expected_name
        assert str(primitive_element) == expected_name
        assert isinstance(primitive_element.value, partial)
        assert primitive_element.value.args == expected_partial_element.args
        assert primitive_element.value.keywords == expected_partial_element.keywords

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_primitive_element_mapping(self):
        """
        <b>Description:</b>
        Check "PrimitiveElementMapping" Enum class elements

        <b>Expected results:</b>
        Test passes if "PrimitiveElementMapping" Enum class length is equal to expected value and its elements have
        expected "value" and "name" attributes and value returned by "__str__" method
        """

        assert len(PrimitiveElementMapping) == 5
        self.check_primitive_element(
            primitive_element=PrimitiveElementMapping.INTEGER,
            expected_name="INTEGER",
            expected_partial_element=partial(configurable_integer),
        )
        self.check_primitive_element(
            primitive_element=PrimitiveElementMapping.FLOAT,
            expected_name="FLOAT",
            expected_partial_element=partial(configurable_float),
        )
        self.check_primitive_element(
            primitive_element=PrimitiveElementMapping.BOOLEAN,
            expected_name="BOOLEAN",
            expected_partial_element=partial(configurable_boolean),
        )
        self.check_primitive_element(
            primitive_element=PrimitiveElementMapping.FLOAT_SELECTABLE,
            expected_name="FLOAT_SELECTABLE",
            expected_partial_element=partial(float_selectable),
        )
        self.check_primitive_element(
            primitive_element=PrimitiveElementMapping.SELECTABLE,
            expected_name="SELECTABLE",
            expected_partial_element=partial(selectable),
        )


def check_mapping_element(mapping_element, expected_name, expected_value):
    assert mapping_element.name == expected_name
    assert mapping_element.value == expected_value
    assert str(mapping_element) == expected_name


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestGroupElementMapping:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_group_element_mapping(self):
        """
        <b>Description:</b>
        Check "GroupElementMapping" Enum class elements

        <b>Expected results:</b>
        Test passes if "GroupElementMapping" Enum class length is equal to expected value and its elements have
        expected "name" attribute and value returned by "__str__" method
        """
        assert len(GroupElementMapping) == 2
        check_mapping_element(
            mapping_element=GroupElementMapping.PARAMETER_GROUP,
            expected_name="PARAMETER_GROUP",
            expected_value=ParameterGroup,
        )
        check_mapping_element(
            mapping_element=GroupElementMapping.CONFIGURABLE_PARAMETERS,
            expected_name="CONFIGURABLE_PARAMETERS",
            expected_value=ConfigurableParameters,
        )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestRuleElementMapping:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_group_element_mapping(self):
        """
        <b>Description:</b>
        Check "RuleElementMapping" Enum class elements

        <b>Expected results:</b>
        Test passes if "RuleElementMapping" Enum class length is equal to expected value and its elements have
        expected "name" attribute and value returned by "__str__" method
        """
        assert len(RuleElementMapping) == 2
        check_mapping_element(
            mapping_element=RuleElementMapping.RULE,
            expected_name="RULE",
            expected_value=Rule,
        )
        check_mapping_element(
            mapping_element=RuleElementMapping.UI_RULES,
            expected_name="UI_RULES",
            expected_value=UIRules,
        )



"""DetCon head to get pixel-wise contrastive loss with predictor head."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=unused-argument
from typing import Any, Dict

import torch
from mmseg.models.builder import HEADS, build_loss, build_neck
from torch import nn


@HEADS.register_module()
class DetConHead(nn.Module):
    """DetCon head for pixel-wise contrastive learning.

    Args:
        predictor (dict): configurations for predictor.
        loss_cfg (dict): configurations for detcon loss.
    """

    def __init__(self, predictor: Dict[str, Any], loss_cfg: Dict[str, Any], **kwargs):
        super().__init__()
        self.predictor = build_neck(predictor)
        self.detcon_loss = build_loss(loss_cfg)

    def init_weights(self):
        """Initialize predictor weights."""
        self.predictor.init_weights()

    # pylint: disable=too-many-locals
    def forward(
        self,
        projs: torch.Tensor,
        projs_tgt: torch.Tensor,
        ids: torch.Tensor,
        ids_tgt: torch.Tensor,
        batch_size: int,
        num_samples: int,
    ) -> Dict[str, torch.Tensor]:
        """Forward head.

        Args:
            projs (Tensor): NxC input features.
            projs_tgt (Tensor): NxC target features.
            ids (Tensor): NxC input ids.
            ids_tgt (Tensor): NxC target ids.
            batch_size (int): Batch size to split concatenated features.
            num_samples (int): The number of samples to be sampled.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        preds = self.predictor(projs)
        pred1, pred2 = torch.split(preds.reshape((-1, num_samples, preds.shape[-1])), batch_size)
        id1, id2 = torch.split(ids, batch_size)
        proj1_tgt, proj2_tgt = torch.split(projs_tgt.reshape((-1, num_samples, projs_tgt.shape[-1])), batch_size)
        id1_tgt, id2_tgt = torch.split(ids_tgt, batch_size)

        loss = self.detcon_loss(
            pred1=pred1,
            pred2=pred2,
            target1=proj1_tgt,
            target2=proj2_tgt,
            pind1=id1,
            pind2=id2,
            tind1=id1_tgt,
            tind2=id2_tgt,
        )

        return dict(loss=loss)



"""
Common functions for input parameters validation tests
"""

import numpy as np
import pytest

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset


def check_value_error_exception_raised(correct_parameters: dict, unexpected_values: list, class_or_function) -> None:
    """
    Function checks that ValueError exception is raised when unexpected type values are specified as parameters for
    methods or functions
    """
    for key, value in unexpected_values:
        incorrect_parameters_dict = dict(correct_parameters)
        incorrect_parameters_dict[key] = value
        with pytest.raises(ValueError):
            class_or_function(**incorrect_parameters_dict)


def load_test_dataset():
    """Function prepares DatasetEntity object and labels list"""

    def gen_image(resolution, x1, y1, x2, y2):
        width, height = resolution
        image = np.full([height, width, 3], fill_value=255, dtype=np.uint8)
        image[int(y1 * height) : int(y2 * height), int(x1 * width) : int(x2 * width), :] = np.array(
            [0, 128, 128], dtype=np.uint8
        )[None, None, :]
        return image, Rectangle(x1=x1, y1=y1, x2=x2, y2=y2)

    images = [
        gen_image((640, 480), 0.0, 0.0, 0.5, 0.5),
        gen_image((640, 480), 0.5, 0.0, 1.0, 0.5),
        gen_image((640, 480), 0.0, 0.5, 0.5, 1.0),
        gen_image((640, 480), 0.5, 0.5, 1.0, 1.0),
    ]
    labels = [LabelEntity(name="rect", domain=Domain.DETECTION, id=ID("0"))]

    def get_image(i, subset):
        image, bbox = images[i]
        return DatasetItemEntity(
            media=Image(data=image),
            annotation_scene=AnnotationSceneEntity(
                annotations=[Annotation(bbox, labels=[ScoredLabel(label=labels[0])])],
                kind=AnnotationSceneKind.ANNOTATION,
            ),
            subset=subset,
        )

    items = [
        get_image(0, Subset.TRAINING),
        get_image(1, Subset.TRAINING),
        get_image(2, Subset.TRAINING),
        get_image(3, Subset.TRAINING),
        get_image(0, Subset.TRAINING),
        get_image(1, Subset.TRAINING),
        get_image(2, Subset.TRAINING),
        get_image(3, Subset.TRAINING),
        get_image(0, Subset.TRAINING),
        get_image(1, Subset.TRAINING),
        get_image(0, Subset.VALIDATION),
        get_image(1, Subset.VALIDATION),
        get_image(2, Subset.VALIDATION),
        get_image(3, Subset.VALIDATION),
        get_image(0, Subset.TESTING),
        get_image(1, Subset.TESTING),
        get_image(2, Subset.TESTING),
        get_image(3, Subset.TESTING),
    ]
    return DatasetEntity(items), labels



"""Module for otx.core.ov.graph."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

# TODO: Need to remove comment with ignore mypy and fix mypy issues
from .graph import Graph  # type: ignore[attr-defined]

__all__ = ["Graph"]



"""Tests for Action Classification with OTX CLI."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import json
import os
from pathlib import Path
from timeit import default_timer as timer

import pytest

from otx.cli.registry import Registry
from tests.regression.regression_test_helpers import (
    REGRESSION_TEST_EPOCHS,
    TIME_LOG,
    get_result_dict,
    get_template_performance,
    load_regression_configuration,
)
from tests.test_suite.e2e_test_system import e2e_pytest_component
from tests.test_suite.run_test_command import (
    otx_export_testing,
    otx_train_testing,
    pot_optimize_testing,
)

from tests.regression.regression_command import (
    regression_eval_testing,
    regression_openvino_testing,
    regression_deployment_testing,
    regression_nncf_eval_testing,
    regression_pot_eval_testing,
    regression_train_time_testing,
    regression_eval_time_testing,
)

# Configurations for regression test.
TASK_TYPE = "action_classification"
TRAIN_TYPE = "supervised"
LABEL_TYPE = "multi_class"

otx_dir = os.getcwd()
templates = Registry("otx/algorithms/action").filter(task_type=TASK_TYPE.upper()).templates
templates_ids = [template.model_template_id for template in templates]


result_dict = get_result_dict(TASK_TYPE)
result_dir = f"/tmp/regression_test_results/{TASK_TYPE}"
Path(result_dir).mkdir(parents=True, exist_ok=True)

action_cls_regression_config = load_regression_configuration(otx_dir, TASK_TYPE, TRAIN_TYPE, LABEL_TYPE)
action_cls_data_args = action_cls_regression_config["data_path"]
action_cls_data_args["train_params"] = ["params", "--learning_parameters.num_iters", REGRESSION_TEST_EPOCHS]


class TestRegressionActionClassification:
    def setup_method(self):
        self.label_type = LABEL_TYPE
        self.performance = {}

    def teardown(self):
        with open(f"{result_dir}/result.json", "w") as result_file:
            json.dump(result_dict, result_file, indent=4)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_train(self, template, tmp_dir_path):
        self.performance[template.name] = {}

        tmp_dir_path = tmp_dir_path / TASK_TYPE
        train_start_time = timer()
        otx_train_testing(template, tmp_dir_path, otx_dir, action_cls_data_args)
        train_elapsed_time = timer() - train_start_time

        infer_start_time = timer()
        test_result = regression_eval_testing(
            template,
            tmp_dir_path,
            otx_dir,
            action_cls_data_args,
            action_cls_regression_config["regression_criteria"]["train"],
            self.performance[template.name],
        )
        infer_elapsed_time = timer() - infer_start_time

        self.performance[template.name][TIME_LOG["train_time"]] = round(train_elapsed_time, 3)
        self.performance[template.name][TIME_LOG["infer_time"]] = round(infer_elapsed_time, 3)
        result_dict[TASK_TYPE][LABEL_TYPE][TRAIN_TYPE]["train"].append(self.performance)

        assert test_result["passed"] is True, test_result["log"]

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_train_kpi_test(self, template):
        results = result_dict[TASK_TYPE][self.label_type][TRAIN_TYPE]["train"]
        performance = get_template_performance(results, template)

        kpi_train_result = regression_train_time_testing(
            train_time_criteria=action_cls_regression_config["kpi_e2e_train_time_criteria"]["train"],
            e2e_train_time=performance[template.name][TIME_LOG["train_time"]],
            template=template,
        )

        kpi_eval_result = regression_eval_time_testing(
            eval_time_criteria=action_cls_regression_config["kpi_e2e_eval_time_criteria"]["train"],
            e2e_eval_time=performance[template.name][TIME_LOG["infer_time"]],
            template=template,
        )

        assert kpi_train_result["passed"] is True, kpi_train_result["log"]
        assert kpi_eval_result["passed"] is True, kpi_eval_result["log"]

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_export_eval_openvino(self, template, tmp_dir_path):
        if template.name == "MoViNet":
            pytest.skip(reason="[CVS-106939] MoViNet export --> eval issue.")
        self.performance[template.name] = {}

        tmp_dir_path = tmp_dir_path / TASK_TYPE
        export_start_time = timer()
        otx_export_testing(template, tmp_dir_path)
        export_elapsed_time = timer() - export_start_time

        export_eval_start_time = timer()
        test_result = regression_openvino_testing(
            template,
            tmp_dir_path,
            otx_dir,
            action_cls_data_args,
            threshold=0.05,
            criteria=action_cls_regression_config["regression_criteria"]["export"],
            reg_threshold=0.10,
            result_dict=self.performance[template.name],
        )
        export_eval_elapsed_time = timer() - export_eval_start_time

        self.performance[template.name][TIME_LOG["export_time"]] = round(export_elapsed_time, 3)
        self.performance[template.name][TIME_LOG["export_eval_time"]] = round(export_eval_elapsed_time, 3)
        result_dict[TASK_TYPE][self.label_type][TRAIN_TYPE]["export"].append(self.performance)

        assert test_result["passed"] is True, test_result["log"]

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_optimize_eval(self, template, tmp_dir_path):
        self.performance[template.name] = {}

        tmp_dir_path = tmp_dir_path / TASK_TYPE
        pot_start_time = timer()
        pot_optimize_testing(template, tmp_dir_path, otx_dir, action_cls_data_args)
        pot_elapsed_time = timer() - pot_start_time

        pot_eval_start_time = timer()
        test_result = regression_pot_eval_testing(
            template,
            tmp_dir_path,
            otx_dir,
            action_cls_data_args,
            criteria=action_cls_regression_config["regression_criteria"]["pot"],
            reg_threshold=0.10,
            result_dict=self.performance[template.name],
        )
        pot_eval_elapsed_time = timer() - pot_eval_start_time

        self.performance[template.name][TIME_LOG["pot_time"]] = round(pot_elapsed_time, 3)
        self.performance[template.name][TIME_LOG["pot_eval_time"]] = round(pot_eval_elapsed_time, 3)
        result_dict[TASK_TYPE][self.label_type][TRAIN_TYPE]["pot"].append(self.performance)

        assert test_result["passed"] is True, test_result["log"]



"""Loss list of mmdetection adapters."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .cross_focal_loss import CrossSigmoidFocalLoss
from .l2sp_loss import L2SPLoss

__all__ = ["CrossSigmoidFocalLoss", "L2SPLoss"]



"""File system related utilities."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

OTX_CACHE = os.path.expanduser(
    os.getenv(
        "OTX_CACHE",
        os.path.join(os.getenv("XDG_CACHE_HOME", "~/.cache"), "otx"),
    )
)
os.makedirs(OTX_CACHE, exist_ok=True)



# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
from pathlib import Path
from typing import Any, Dict, List, Union

from otx.api.entities.model_template import ModelTemplate

TEST_TYPES = ["train", "export", "deploy", "nncf", "pot"]
TASK_TYPES = [
    "classification",
    "detection",
    "semantic_segmentation",
    "instance_segmentation",
    "action_classification",
    "action_detection",
    "anomaly",
]
TRAIN_TYPES = ["supervised", "semi_supervised", "self_supervised", "class_incr", "tiling"]
LABEL_TYPES = ["multi_class", "multi_label", "h_label", "supcon"]

REGRESSION_TEST_EPOCHS = "10"

ANOMALY_DATASET_CATEGORIES = [
    "bottle",
    "cable",
    "capsule",
    "carpet",
    "grid",
    "hazelnut",
    "leather",
    "metal_nut",
    "pill",
    "screw",
    "tile",
    "toothbrush",
    "transistor",
    "wood",
    "zipper",
]


TIME_LOG = {
    "train_time": "Train + val time (sec.)",
    "infer_time": "Infer time (sec.)",
    "export_time": "Export time (sec.)",
    "export_eval_time": "Export eval time (sec.)",
    "deploy_time": "Deploy time (sec.)",
    "deploy_eval_time": "Deploy eval time (sec.)",
    "nncf_time": "NNCF time (sec.)",
    "nncf_eval_time": "NNCF eval time (sec.)",
    "pot_time": "POT time (sec.)",
    "pot_eval_time": "POT eval time (sec.)",
}


def get_result_dict(task_type: str) -> Dict[str, Any]:
    result_dict = {task_type: {}}
    if "anomaly" not in task_type:
        for label_type in LABEL_TYPES:
            result_dict[task_type][label_type] = {}
            for train_type in TRAIN_TYPES:
                result_dict[task_type][label_type][train_type] = {}
                for test_type in TEST_TYPES:
                    result_dict[task_type][label_type][train_type][test_type] = []
    else:
        for test_type in TEST_TYPES:
            result_dict[task_type][test_type] = {}
            for category in ANOMALY_DATASET_CATEGORIES:
                result_dict[task_type][test_type][category] = []

    return result_dict


def load_regression_config(otx_dir: str) -> Dict[str, Any]:
    """Load regression config from path.

    Args:
        otx_dir (str): The path of otx root directory

    Returns:
        Dict[str, Any]: The dictionary that includes data roots
    """
    root_path = Path(otx_dir)
    with open(root_path / ("tests/regression/regression_config.json"), "r") as f:
        reg_config = json.load(f)
    return reg_config


def load_regression_configuration(
    otx_dir: str, task_type: str, train_type: str = "", label_type: str = ""
) -> Dict[str, Union[str, int, float]]:
    """Load dataset path according to task, train, label types.

    Args:
        otx_dir (str): The path of otx root directoy
        task_type (str): ["classification", "detection", "instance segmentation", "semantic segmentation",
                            "action_classification", "action_detection", "anomaly"]
        train_type (str): ["supervised", "semi_supervised", "self_supervised", "class_incr"]
        label_type (str): ["multi_class", "multi_label", "h_label", "supcon"]

    Returns:
        Dict[str, Union[int, float]]: The dictionary that includes model criteria
    """
    reg_config = load_regression_config(otx_dir)
    result: Dict[str, Union[str, int, float]] = {
        "data_path": "",
        "model_criteria": 0,
    }

    if "anomaly" not in task_type:
        if train_type == "" or label_type == "":
            raise ValueError()
        result["regression_criteria"] = reg_config["regression_criteria"][task_type][train_type][label_type]
        result["kpi_e2e_train_time_criteria"] = reg_config["kpi_e2e_train_time_criteria"][task_type][train_type][
            label_type
        ]
        result["kpi_e2e_eval_time_criteria"] = reg_config["kpi_e2e_eval_time_criteria"][task_type][train_type][
            label_type
        ]
        result["data_path"] = reg_config["data_path"][task_type][train_type][label_type]
    else:
        result["regression_criteria"] = reg_config["regression_criteria"][task_type]
        result["kpi_e2e_train_time_criteria"] = reg_config["kpi_e2e_train_time_criteria"][task_type]
        result["kpi_e2e_eval_time_criteria"] = reg_config["kpi_e2e_eval_time_criteria"][task_type]
        result["data_path"] = reg_config["data_path"][task_type]

    return result


def get_template_performance(results: List[Dict], template: ModelTemplate):
    """Get proper template performance inside of performance list."""
    performance = None
    for result in results:
        template_name = list(result.keys())[0]
        if template_name == template.name:
            performance = result
            break
    if performance is None:
        raise ValueError("Performance is None.")
    return performance



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
from typing import List

import numpy as np
import pytest
import torch
from mmcv import Config, ConfigDict
from mmdet.datasets import build_dataloader, build_dataset
from mmdet.models import DETECTORS
from openvino.model_zoo.model_api.adapters import OpenvinoAdapter, create_core
from torch import nn

from otx.algorithms.common.adapters.mmcv.utils.config_utils import MPAConfig
from otx.algorithms.common.adapters.mmdeploy.apis import MMdeployExporter
from otx.algorithms.detection.adapters.mmdet.task import MMDetectionTask
from otx.algorithms.detection.adapters.mmdet.utils import build_detector, patch_tiling
from otx.api.configuration.helper import create
from otx.api.entities.annotation import AnnotationSceneEntity, AnnotationSceneKind
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity, DatasetPurpose
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.usecases.adapters.model_adapter import ModelAdapter
from otx.api.utils.shape_factory import ShapeFactory
from tests.test_helpers import generate_random_annotated_image
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.detection.test_helpers import (
    DEFAULT_ISEG_TEMPLATE_DIR,
    init_environment,
)


@DETECTORS.register_module(force=True)
class MockDetModel(nn.Module):
    def __init__(self, backbone, train_cfg=None, test_cfg=None, init_cfg=None):
        super().__init__()
        self.conv = torch.nn.Conv2d(3, 3, 3)
        self.box_dummy = torch.nn.AdaptiveAvgPool2d((1, 5))
        self.label_dummy = torch.nn.AdaptiveAvgPool2d((1))
        self.mask_dummy = torch.nn.AdaptiveAvgPool2d((28, 28))

    def forward(self, *args, **kwargs):
        img = args[0]
        x = self.conv(img)
        boxes = self.box_dummy(x).mean(1)
        labels = self.label_dummy(x).mean(1)
        masks = self.mask_dummy(x).mean(1)
        return boxes, labels, masks


def create_otx_dataset(height: int, width: int, labels: List[str], domain: Domain = Domain.DETECTION):
    """Create a random OTX dataset.

    Args:
        height (int): The height of the image
        width (int): The width of the image

    Returns:
        DatasetEntity: OTX dataset entity
        List[LabelEntity]: The list of labels
    """
    labels = []
    for label in ["rectangle", "ellipse", "triangle"]:
        labels.append(LabelEntity(name=label, domain=domain))
    image, anno_list = generate_random_annotated_image(width, height, labels)
    image = Image(data=image)
    annotation_scene = AnnotationSceneEntity(annotations=anno_list, kind=AnnotationSceneKind.ANNOTATION)
    dataset_item = DatasetItemEntity(media=image, annotation_scene=annotation_scene)
    return DatasetEntity([dataset_item]), labels


class TestTilingDetection:
    """Test the tiling detection algorithm."""

    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        """Setup the test case."""
        self.height = 1024
        self.width = 1024
        self.label_names = ["rectangle", "ellipse", "triangle"]
        self.tile_cfg = dict(
            tile_size=np.random.randint(low=100, high=500),
            overlap_ratio=np.random.uniform(low=0.0, high=0.5),
            max_per_img=np.random.randint(low=1, high=10000),
            max_annotation=1000,
        )
        self.dataloader_cfg = dict(samples_per_gpu=1, workers_per_gpu=1)
        self.otx_dataset, self.labels = create_otx_dataset(self.height, self.width, self.label_names)

        img_norm_cfg = dict(mean=[0, 0, 0], std=[255, 255, 255], to_rgb=True)

        self.train_data_cfg = ConfigDict(
            dict(
                type="ImageTilingDataset",
                pipeline=[
                    dict(type="Resize", img_scale=(self.height, self.width), keep_ratio=False),
                    dict(type="RandomFlip", flip_ratio=0.5),
                    dict(type="Pad", size_divisor=32),
                    dict(type="DefaultFormatBundle"),
                    dict(type="Collect", keys=["img", "gt_bboxes", "gt_labels"]),
                ],
                dataset=dict(
                    type="OTXDetDataset",
                    pipeline=[
                        dict(type="LoadImageFromOTXDataset"),
                        dict(
                            type="LoadAnnotationFromOTXDataset",
                            with_bbox=True,
                            with_mask=False,
                            domain=Domain.DETECTION,
                            min_size=-1,
                        ),
                    ],
                    otx_dataset=self.otx_dataset,
                    labels=self.labels,
                    domain=Domain.DETECTION,
                ),
                **self.tile_cfg
            )
        )

        self.test_data_cfg = ConfigDict(
            dict(
                type="ImageTilingDataset",
                pipeline=[
                    dict(
                        type="MultiScaleFlipAug",
                        img_scale=(self.height, self.width),
                        flip=False,
                        transforms=[
                            dict(type="Resize", keep_ratio=False),
                            dict(type="Normalize", **img_norm_cfg),
                            dict(type="ImageToTensor", keys=["img"]),
                            dict(type="Collect", keys=["img"]),
                        ],
                    )
                ],
                dataset=dict(
                    type="OTXDetDataset",
                    pipeline=[dict(type="LoadImageFromOTXDataset")],
                    otx_dataset=self.otx_dataset.with_empty_annotations(),
                    labels=list(self.labels),
                    domain=Domain.DETECTION,
                ),
                test_mode=True,
                **self.tile_cfg
            )
        )

    @e2e_pytest_unit
    def test_tiling_train_dataloader(self):
        """Test that the training dataloader is built correctly for tiling."""

        dataset = build_dataset(self.train_data_cfg)
        train_dataloader = build_dataloader(dataset, **self.dataloader_cfg)
        for data in train_dataloader:
            assert isinstance(data["img"].data[0], torch.Tensor)
            assert isinstance(data["gt_bboxes"].data[0][0], torch.Tensor)
            assert isinstance(data["gt_labels"].data[0][0], torch.Tensor)

    @e2e_pytest_unit
    def test_tiling_test_dataloader(self):
        """Test that the testing dataloader is built correctly for tiling."""

        dataset = build_dataset(self.test_data_cfg)
        stride = int((1 - self.tile_cfg["overlap_ratio"]) * self.tile_cfg["tile_size"])
        num_tile_rows = ((self.height - self.tile_cfg["tile_size"]) // stride) + 1
        num_tile_cols = ((self.width - self.tile_cfg["tile_size"]) // stride) + 1
        # +1 for the original image
        assert len(dataset) == (num_tile_rows * num_tile_cols) + 1, "Incorrect number of tiles"

        test_dataloader = build_dataloader(dataset, **self.dataloader_cfg)
        for data in test_dataloader:
            assert isinstance(data["img"][0], torch.Tensor)
            assert "gt_bboxes" not in data
            assert "gt_labels" not in data

    @e2e_pytest_unit
    def test_inference_merge(self):
        """Test that the inference merge works correctly."""
        dataset = build_dataset(self.test_data_cfg)

        # create simulated inference results
        results: List[List[np.ndarray]] = []
        for i in range(len(dataset)):
            results.append([])
            for _ in range(len(self.labels)):
                results[i].append(np.zeros((0, 5), dtype=np.float32))

        # generate tile predictions
        for i in range(len(dataset)):
            img_width, img_height = self.tile_cfg["tile_size"], self.tile_cfg["tile_size"]
            if i == 0:
                # first index belongs is the full image
                img_width, img_height = self.width, self.height

            _, anno_list = generate_random_annotated_image(img_width, img_height, self.labels)
            for anno in anno_list:
                shape = ShapeFactory.shape_as_rectangle(anno.shape)
                bbox = np.array([shape.x1, shape.y1, shape.x2, shape.y2], np.float32)
                bbox *= np.tile([img_width, img_height], 2)
                score_bbox = np.array([*bbox, np.random.rand()], np.float32)
                label_idx = self.label_names.index(anno.get_labels()[0].name)
                results[i][label_idx] = np.append(results[i][label_idx], [score_bbox], axis=0)

        merged_bbox_results = dataset.merge(results)
        assert len(merged_bbox_results) == dataset.num_samples

    @e2e_pytest_unit
    def test_load_tiling_parameters(self, tmp_dir_path):
        maskrcnn_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_ISEG_TEMPLATE_DIR, "model.py"))
        detector = build_detector(maskrcnn_cfg)

        # Enable tiling and save weights
        model_template = parse_model_template(os.path.join(DEFAULT_ISEG_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        hyper_parameters.tiling_parameters.enable_tiling = True
        task_env = init_environment(hyper_parameters, model_template)
        output_model = ModelEntity(self.otx_dataset, task_env.get_model_configuration())
        task = MMDetectionTask(task_env, output_path=str(tmp_dir_path))
        model_ckpt = os.path.join(tmp_dir_path, "maskrcnn.pth")
        torch.save(detector.state_dict(), model_ckpt)
        task._model_ckpt = model_ckpt
        task.save_model(output_model)
        for filename, model_adapter in output_model.model_adapters.items():
            with open(os.path.join(tmp_dir_path, filename), "wb") as write_file:
                write_file.write(model_adapter.data)

        # Read tiling parameters from weights
        with open(os.path.join(tmp_dir_path, "weights.pth"), "rb") as f:
            bin_data = f.read()
            model = ModelEntity(
                self.otx_dataset,
                configuration=task_env.get_model_configuration(),
                model_adapters={"weights.pth": ModelAdapter(bin_data)},
            )
        task_env.model = model
        task = MMDetectionTask(task_env, output_path=str(tmp_dir_path))

    @e2e_pytest_unit
    def test_patch_tiling_func(self):
        """Test that patch_tiling function works correctly."""
        cfg = MPAConfig.fromfile(os.path.join(DEFAULT_ISEG_TEMPLATE_DIR, "model.py"))
        model_template = parse_model_template(os.path.join(DEFAULT_ISEG_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        hyper_parameters.tiling_parameters.enable_tiling = True

        self.otx_dataset.purpose = DatasetPurpose.TRAINING
        patch_tiling(cfg, hyper_parameters, self.otx_dataset)

        self.otx_dataset.purpose = DatasetPurpose.INFERENCE
        patch_tiling(cfg, hyper_parameters, self.otx_dataset)

    @e2e_pytest_unit
    @pytest.mark.parametrize("scale_factor", [1, 1.5, 2, 3, 4])
    def test_tile_ir_scale_deploy(self, tmp_dir_path, scale_factor):
        """Test that the IR scale factor is correctly applied during inference."""
        model_template = parse_model_template(os.path.join(DEFAULT_ISEG_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        hyper_parameters.tiling_parameters.enable_tiling = True
        hyper_parameters.tiling_parameters.tile_ir_scale_factor = scale_factor
        task_env = init_environment(hyper_parameters, model_template)
        img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
        task = MMDetectionTask(task_env)
        pipeline = [
            dict(type="LoadImageFromFile"),
            dict(
                type="MultiScaleFlipAug",
                img_scale=(800, 800),
                flip=False,
                transforms=[
                    dict(type="Resize", keep_ratio=False),
                    dict(type="RandomFlip"),
                    dict(type="Normalize", **img_norm_cfg),
                    dict(type="Pad", size_divisor=32),
                    dict(type="DefaultFormatBundle"),
                    dict(type="Collect", keys=["img"]),
                ],
            ),
        ]
        config = Config(
            dict(model=dict(type="MockDetModel", backbone=dict(init_cfg=None)), data=dict(test=dict(pipeline=pipeline)))
        )

        deploy_cfg = task._init_deploy_cfg(config)
        onnx_path = MMdeployExporter.torch2onnx(
            tmp_dir_path,
            np.zeros((50, 50, 3), dtype=np.float32),
            config,
            deploy_cfg,
        )
        assert isinstance(onnx_path, str)
        assert os.path.exists(onnx_path)

        openvino_paths = MMdeployExporter.onnx2openvino(
            tmp_dir_path,
            onnx_path,
            deploy_cfg,
        )
        for openvino_path in openvino_paths:
            assert os.path.exists(openvino_path)

        task._init_task()
        original_width, original_height = task._recipe_cfg.data.test.pipeline[0].img_scale  # w, h

        model_adapter = OpenvinoAdapter(create_core(), openvino_paths[0], openvino_paths[1])

        ir_input_shape = model_adapter.get_input_layers()["image"].shape
        _, _, ir_height, ir_width = ir_input_shape
        assert ir_height == original_height * scale_factor
        assert ir_width == original_width * scale_factor

    @e2e_pytest_unit
    def test_max_annotation(self, max_annotation=200):
        otx_dataset, labels = create_otx_dataset(
            self.height, self.width, self.label_names, Domain.INSTANCE_SEGMENTATION
        )
        factor = int(max_annotation / len(otx_dataset[0].annotation_scene.annotations)) + 2
        otx_dataset[0].annotation_scene.annotations = otx_dataset[0].annotation_scene.annotations * factor
        dataloader_cfg = dict(samples_per_gpu=1, workers_per_gpu=1)

        tile_cfg = dict(
            tile_size=np.random.randint(low=100, high=500),
            overlap_ratio=np.random.uniform(low=0.0, high=0.5),
            max_per_img=np.random.randint(low=1, high=10000),
            max_annotation=max_annotation,
        )
        train_data_cfg = ConfigDict(
            dict(
                type="ImageTilingDataset",
                pipeline=[
                    dict(type="Resize", img_scale=(self.height, self.width), keep_ratio=False),
                    dict(type="RandomFlip", flip_ratio=0.5),
                    dict(type="Pad", size_divisor=32),
                    dict(type="DefaultFormatBundle"),
                    dict(type="Collect", keys=["img", "gt_bboxes", "gt_labels", "gt_masks"]),
                ],
                dataset=dict(
                    type="OTXDetDataset",
                    pipeline=[
                        dict(type="LoadImageFromOTXDataset"),
                        dict(
                            type="LoadAnnotationFromOTXDataset",
                            with_bbox=True,
                            with_mask=True,
                            domain=Domain.INSTANCE_SEGMENTATION,
                            min_size=-1,
                        ),
                    ],
                    otx_dataset=otx_dataset,
                    labels=labels,
                    domain=Domain.INSTANCE_SEGMENTATION,
                ),
                **tile_cfg
            )
        )

        # original annotation over the limitation
        assert len(otx_dataset[0].annotation_scene.annotations) > max_annotation
        dataset = build_dataset(train_data_cfg)
        train_dataloader = build_dataloader(dataset, **dataloader_cfg)
        # check gt annotation is under the limitation
        for data in train_dataloader:
            assert len(data["gt_bboxes"].data[0][0]) <= max_annotation
            assert len(data["gt_labels"].data[0][0]) <= max_annotation
            assert len(data["gt_masks"].data[0][0]) <= max_annotation



"""OTX Adapters - openvino."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .task import OpenVINOSegmentationInferencer, OpenVINOSegmentationTask, OTXOpenVinoDataLoader

__all__ = ["OpenVINOSegmentationTask", "OpenVINOSegmentationInferencer", "OTXOpenVinoDataLoader"]



"""OTX algorithms.detection.adapters.mmdet.models.roi_heads."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# flake8: noqa
from . import bbox_heads, mask_heads, roi_extractors

__all__ = ["bbox_heads", "mask_heads", "roi_extractors"]



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import pytest
import torch

from otx.core.ov.ops.shape_manipulations import (
    ReshapeV1,
    ShapeOfV0,
    ShapeOfV3,
    SqueezeV0,
    UnsqueezeV0,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestSqueezeV0:
    @e2e_pytest_unit
    def test_forward(self):
        op = SqueezeV0("dummy", shape=(1,))
        input = torch.randn(1, 1, 5, 1, 1)

        output = op(input)
        assert output.shape == (5,)

        output = op(input, torch.tensor(0))
        assert output.shape == (1, 5, 1, 1)

        output = op(input, torch.tensor([0, 1]))
        assert output.shape == (5, 1, 1)

        output = op(input, torch.tensor([-2]))
        assert output.shape == (1, 1, 5, 1)


class TestUnsqueezeV0:
    @e2e_pytest_unit
    def test_forward(self):
        op = UnsqueezeV0("dummy", shape=(1,))
        input = torch.randn(2, 3)

        output = op(input, torch.tensor(2))
        assert output.shape == (2, 3, 1)

        output = op(input, torch.tensor([0, 1]))
        assert output.shape == (1, 2, 1, 3)

        output = op(input, torch.tensor([-1]))
        assert output.shape == (2, 3, 1)

        output = op(input, torch.tensor([-1, -2]))
        assert output.shape == (1, 2, 1, 3)


class TestReshapeV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = ReshapeV1("dummy", shape=(1,), special_zero=False)
        input = torch.randn(2, 5, 5, 0)
        output = op(input, torch.tensor((0, 4)))
        assert output.shape == (0, 4)

        op = ReshapeV1("dummy", shape=(1,), special_zero=True)
        input = torch.randn(2, 5, 5, 24)
        output = op(input, torch.tensor((0, -1, 4)))
        assert output.shape == (2, 150, 4)

        op = ReshapeV1("dummy", shape=(1,), special_zero=True)
        input = torch.randn(2, 2, 3)
        output = op(input, torch.tensor((0, 0, 1, -1)))
        assert output.shape == (2, 2, 1, 3)

        op = ReshapeV1("dummy", shape=(1,), special_zero=True)
        input = torch.randn(3, 1, 1)
        output = op(input, torch.tensor((0, -1)))
        assert output.shape == (3, 1)


class TestShapeOfV0:
    @e2e_pytest_unit
    def test_forward(self):
        op = ShapeOfV0("dummy", shape=(1,))
        assert torch.equal(op(torch.randn(5, 3)), torch.tensor([5, 3]))


class TestShapeOfV3:
    @e2e_pytest_unit
    def test_invalid_attr(self):
        with pytest.raises(ValueError):
            ShapeOfV3("dummy", shape=(1,), output_type="error")

    @e2e_pytest_unit
    def test_forward(self):
        op = ShapeOfV3("dummy", shape=(1,), output_type="i32")
        output = op(torch.randn(5, 3))
        assert torch.equal(output, torch.tensor([5, 3]))
        assert output.dtype == torch.int32

        op = ShapeOfV3("dummy", shape=(1,), output_type="i64")
        output = op(torch.randn(5, 3))
        assert torch.equal(output, torch.tensor([5, 3]))
        assert output.dtype == torch.int64



"""MoViNet head for otx action recognition."""
# Copyright (c) OpenMMLab. All rights reserved.
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from mmaction.models.builder import HEADS
from mmaction.models.heads.base import BaseHead
from mmcv.cnn import normal_init
from torch import nn

from otx.algorithms.action.adapters.mmaction.models.backbones.movinet import ConvBlock3D


@HEADS.register_module()
class MoViNetHead(BaseHead):
    """Classification head for MoViNet.

    Args:
        num_classes (int): Number of classes to be classified.
        in_channels (int): Number of channels in input feature.
        hidden_dim (int): Number of channels in hidden layer.
        tf_like (bool): If True, uses TensorFlow-style padding. Default: False.
        conv_type (str): Type of convolutional layer. Default: '3d'.
        loss_cls (dict): Config for building loss. Default: dict(type='CrossEntropyLoss').
        spatial_type (str): Pooling type in spatial dimension. Default: 'avg'.
        dropout_ratio (float): Probability of dropout layer. Default: 0.5.
        init_std (float): Standard deviation for initialization. Default: 0.1.
    """

    def __init__(
        self,
        num_classes: int,
        in_channels: int,
        hidden_dim: int,
        loss_cls: dict,
        tf_like: bool = False,
        conv_type: str = "3d",
    ):
        super().__init__(num_classes, in_channels, loss_cls)
        self.init_std = 0.1
        self.classifier = nn.Sequential(
            ConvBlock3D(
                in_channels,
                hidden_dim,
                kernel_size=(1, 1, 1),
                tf_like=tf_like,
                conv_type=conv_type,
                bias=True,
            ),
            nn.SiLU(),
            nn.Dropout(p=0.2, inplace=True),
            ConvBlock3D(
                hidden_dim,
                num_classes,
                kernel_size=(1, 1, 1),
                tf_like=tf_like,
                conv_type=conv_type,
                bias=True,
            ),
        )

    def init_weights(self):
        """Initialize the parameters from scratch."""
        normal_init(self.classifier, std=self.init_std)

    def forward(self, x):
        """Defines the computation performed at every call.

        Args:
            x (torch.Tensor): The input data.

        Returns:
            torch.Tensor: The classification scores for input samples.
        """
        # [N, in_channels, T, H, W]
        cls_score = self.classifier(x)
        cls_score = cls_score.flatten(1)
        # [N, num_classes]
        return cls_score



"""OTX Algorithms - Segmentation."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

MMSEG_AVAILABLE = True

try:
    import mmseg  # noqa: F401
except ImportError:
    MMSEG_AVAILABLE = False
